unit convert;

interface

uses Windows,SysUtils,Classes,resource;

const UNKNOWN_FORMAT = -1;
      CAN_CREATE = -1;

function FromRes(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;var Ext:string;ResType,ResName:PChar;ResLang:word):integer;
function ToRes(Source:pointer;SourceSize:integer;var Dest:pointer;Ext:string;ResType,ResName:PChar;ResLang:word):integer;

implementation

function RemoveSpaces(const s:string):string;
var i,n:integer;
begin
  i:=1; n:=length(s);
  while (i<length(s)) and ((s[i]=' ') or (s[i]=#9)) do inc(i);
  while (n>=i) and ((s[n]=' ') or (s[n]=#9)) do dec(n);
  result:=copy(s,i,n-i+1);
end;

function RemoveTabs(s:string):string;
var i:integer;
begin
  result:=s;
  for i:=1 to length(result) do if result[i]=#9 then result[i]:=' ';
end;

function IsDec(s:string):boolean;
var i:integer;
begin
  if (s='') or (s='-') or (s='+') then result:=false
  else begin
    result:=true;
    if (s[1]='-') or (s[1]='+') then i:=2 else i:=1;
    while (i<=length(s)) and result do begin
      result:=(s[i]>='0') and (s[i]<='9');
      inc(i);
    end;
  end;
end;

function TestFlag(Value,Flag:integer):boolean;
begin
  result:=(Value and Flag)=Flag;
end;

function SortByObject(List:TstringList;Index1,Index2:integer):integer;
begin
  result:=NativeInt(List.Objects[Index1])-NativeInt(List.Objects[Index2]);
end;

function IsHex(s:string):boolean;
var i:integer;
begin
  if s='' then result:=false
  else begin
    result:=true;
    i:=1;
    while (i<=length(s)) and result do begin
      result:=((s[i]>='0') and (s[i]<='9')) or ((s[i]>='A') and (s[i]<='F')) or ((s[i]>='a') or (s[i]<='f'));
      inc(i);
    end;
  end;
end;

function NextPos(Substr,S:string;StartPos:integer):integer;
begin
  result:=StartPos;
  while (result<length(S)-length(Substr)+1) and (Substr<>copy(S,result,length(Substr))) do inc(result);
  if result>=length(S)-length(Substr)+1 then result:=0;
end;

function UpperChar(c:char):char;
begin
  result:=char(integer(CharUpper(PChar(integer(c)))));
end;

function LanguageString(LangID:word):string;
var p,s:word;
begin
  p:=LangID and $03FF;
  s:=(LangID and $FC00) shr 10;
  result:='LANGUAGE '+IntToStr(p)+', '+IntToStr(s);
end;

function AddString(Text:TStrings;CurrentSize:integer;S:string):integer;
begin
  result:=CurrentSize+length(S)+2;
  if Assigned(Text) then Text.Add(S);
end;

function HexToInt(Hex:string):integer;
var i,n:integer;
begin
  result:=0;
  i:=length(Hex);
  n:=1;
  while i>0 do begin
    case Hex[i] of
      '0'..'9':result:=result+n*(ord(Hex[i])-ord('0'));
      'A'..'F':result:=result+n*(ord(Hex[i])-ord('A')+10);
      'a'..'f':result:=result+n*(ord(Hex[i])-ord('a')+10);
      else raise EConvertError.Create(Hex+' is not valid hexadicimal number');
    end;
    dec(i);
    n:=n*16;
  end;
end;

function ShiftPointer(p:pointer;ofst:integer):pointer;
begin
  {
  asm
    mov eax,p
    add eax,ofst
    mov result,eax
  end;
  }
  Result := Pointer(NativeUInt(p) + ofst);
end;

function CStyleToStr(s:string):string;
var i:integer;
begin
  i:=1;
  result:='';
  while (i<=length(s)) do begin
    if (s[i]<>'\') or (i=length(s)) then result:=result+s[i]
    else begin
      case UpCase(s[i+1]) of
        '"':result:=result+'"';
        'N':result:=result+#10;
        'R':result:=result+#13;
        'T':result:=result+#9;
        'O':result:=result+#8;
        'X':if (i<length(s)-4) and IsHex(copy(s,i+2,3)) then begin
          result:=result+chr(HexToInt(copy(s,i+2,3)));
          i:=i+3;
        end;
        '\':result:=result+'\';
        else result:=result+'\'+s[i+1];
      end;
      inc(i);
    end;
    inc(i);
  end;
end;

function StrToCStyle(s:string):string;
var i:integer;
begin
  result:='';
  for i:=1 to length(s) do case s[i] of
    #0..#7,#11,#12,#14..#31:result:=result+'\x'+IntToHex(ord(s[i]),3);
    #8:result:=result+'\o';
    #9:result:=result+'\t';
    #10:result:=result+'\n';
    #13:result:=result+'\r';
    '"':result:=result+'\"';
    '\':result:=result+'\\';
    else result:=result+s[i];
  end;
end;

{****    ()  ***********************************************}

type TElement = record
       LineID,Info:integer;
     end;

const EL_KEYWORD   = $00000001; // "K"
      EL_TEXT      = $00000002; // "T"
      EL_CONSTANT  = $00000004; // "C"
      EL_NAME      = $00000008; // "N"
      EL_SEPARATOR = $00000010; // "S"

      EL_TEXT_OR_CONST = EL_TEXT or EL_CONSTANT;
      EL_NAME_OR_CONST = EL_NAME or EL_CONSTANT;

const KeyWordsCount = 17;
      ConstantsCount = 10;

const KeyWords:array[0..KeyWordsCount-1] of string=('STRINGTABLE','SEPARATOR',
      'MENUITEM','LANGUAGE','DIALOGEX','EXSTYLE','CAPTION','CONTROL','DIALOG',
      'MENUEX','POPUP','BEGIN','CLASS','STYLE','FONT','MENU','END');

const Constants:array[0..ConstantsCount-1] of string=('MENUBARBREAK',
      'MENUBREAK','SCROLLBAR','COMBOBOX','CHECKED','GRAYED','STATIC','BUTTON',
      'EDIT','HELP');

const ConstantsValue:array[0..ConstantsCount-1] of integer=(MF_MENUBARBREAK,
      MF_MENUBREAK,$0084,$0085,MFS_CHECKED,MFS_GRAYED,$0082,$0080,$0081,MF_HELP);


function BackCount(s:string;c:char;i:integer):integer;
begin
  result:=0;
  while (i>0) and (s[i]=c) do begin
    inc(result);
    dec(i);
  end;
end;

function UpdateString(s:string):string;
var i,n:integer;
    q:boolean;
begin
//   
  for i:=1 to length(s) do if s[i]=#9 then s[i]:=' ';
//          
  i:=1; n:=length(s);
  while (i<n) and (s[i]=' ') do inc(i);
  while (n>=i) and (s[n]=' ') do dec(n);
  s:=copy(s,i,n-i+1);
//  :   ,     "|",     
  result:='';
  q:=false;
  for i:=1 to length(s) do begin
    if s[i]='"' then begin
      if (not q) or (BackCount(s,'\',i-1) mod 2=0) then q:=not q;
      result:=result+s[i];
    end
    else begin
      if q then result:=result+s[i]
      else case s[i] of
        ' ':if not (result[length(result)] in [',','|',' ',':']) then result:=result+' ';
        ',','|':if result[length(result)]=' ' then result[length(result)]:=s[i] else result:=result+s[i];
        else if q then result:=result+s[i] else result:=result+UpperChar(s[i]);
      end;
    end;
  end;
end;

function NonQPos(substr,s:string):integer;
var q:boolean;
    i,n:integer;
begin
  n:=0;
  repeat;
    result:=NextPos(substr,s,n+1);
    n:=result;
    if result>0 then begin
      i:=1; q:=false;
      while i<result do begin
        if (s[i]='"') and (BackCount(s,'\',i-1) mod 2=0) then q:=not q;
        inc(i);
      end;
      if q then result:=0;
    end;
  until result=n;
end;

function ConstantToInt(S:string;var Constant:integer):boolean;
var i,n:integer;
    st:string;
begin
  Constant:=0; n:=1; result:=true;
  repeat;
    i:=NextPos('|',S,n);
    if i=0 then i:=length(s)+1;
    try
      if (i-n>1) and (S[n]='0') and (S[n+1]='X') then Constant:=Constant or HexToInt(copy(S,n+2,i-n-2))
      else if ((S[n]>='0') and (S[n]<='9')) or (S[n]='-') then Constant:=Constant or StrToInt(copy(S,n,i-n))
      else begin
        st:=copy(S,n,i-n); n:=0;
        while (n<ConstantsCount) and (Constants[n]<>st) do inc(n);
        if n=ConstantsCount then raise EConvertError.Create('"'+st+'" is not valid constant name or number value.');
        Constant:=Constant or ConstantsValue[n];
      end;
    except
      result:=false;
    end;
    n:=i+1;
  until (i>length(s)) or (not result);
end;

function KeyWordToInt(s:string;var k:integer):boolean;
begin
  result:=true; k:=0;
  while (k<KeyWordsCount) and (s<>KeyWords[k]) do inc(k);
  if k=KeyWordsCount then result:=false;
end;

function LoadResource(Text:pointer;TextSize:integer;Strings:TStringList):TMemoryStream;
var EndPointer:pointer;
    Element:TElement;
    q,Comment,Error:boolean;
    i,n,v,Line:integer;
    s,s1:string;
begin
  Error:=true;
  result:=TMemoryStream.Create;
  try
    Strings.Clear; Strings.Sorted:=true;
    EndPointer:=ShiftPointer(Text,TextSize);
    Comment:=false; Line:=0;
    repeat
      s:=''; inc(Line);
      while (Text<>EndPointer) and (char(Text^)<>#10) and (char(Text^)<>#13) do begin
        s:=s+char(Text^);
        inc(NativeUInt(Text));
      end;
      while (Text<>EndPointer) and ((char(Text^)=#10) or (char(Text^)=#13)) do inc(NativeUInt(Text));
      if Comment then begin
        i:=NonQPos('*/',s);
        if i>0 then begin
          s:=copy(s,i+2,length(s)-i-1);
          Comment:=false;
        end
        else s:='';
      end;
      if not Comment then begin
        i:=NonQPos('//',s);
        n:=NonQPos('/*',s);
        if i<>n then begin
          if (i>0) and ((i<n) or (n=0)) then s:=copy(s,1,i-1)
          else begin
            s:=copy(s,1,n-1);
            Comment:=true;
          end;
        end;
      end;
      s:=UpdateString(s);
      if s<>'' then begin
        s:=s+#10; i:=1; n:=1; q:=false; 
        while i<=length(s) do begin
          if s[i]='"' then begin
            if (not q) or (BackCount(s,'\',i-1) mod 2=0) then q:=not q;
          end
          else if (not q) and ((s[i]=',') or (s[i]=' ') or (s[i]=#10)) then begin
            if (s[n]='"') and (s[i-1]='"') then begin
              Element.LineID:=EL_TEXT;
              s1:=CStyleToStr(copy(s,n+1,i-n-2));
              v:=Strings.IndexOf(s1);
              if v=-1 then v:=Strings.AddObject(s1,TObject(Strings.Count));
              Element.Info:=integer(Strings.Objects[v]);
            end
            else if KeyWordToInt(copy(s,n,i-n),Element.Info) then Element.LineID:=EL_KEYWORD
            else begin
              if ConstantToInt(copy(s,n,i-n),Element.Info) then Element.LineID:=EL_CONSTANT
              else begin
                if pos('|',copy(s,n,i-n))<>0 then raise EConvertError.Create('Error at line '+IntToStr(Line)+': "'+copy(s,n,i-n)+'" is not valid constant name or number value.') 
                else begin
                  Element.LineID:=EL_NAME;
                  s1:=copy(s,n,i-n);
                  v:=Strings.indexOf(s1);
                  if v=-1 then v:=Strings.AddObject(s1,TObject(Strings.Count));
                  Element.Info:=integer(Strings.Objects[v]);
                end;
              end;
            end;
            Element.LineID:=Element.LineID or (Line shl 8);
            result.Write(Element,sizeof(Element));
            Element.LineID:=EL_SEPARATOR or (Line shl 8);
            case s[i] of
              ',':Element.Info:=0;
              ' ':Element.Info:=1;
              #10:Element.Info:=2;
            end;
            result.Write(Element,sizeof(Element));
            n:=i+1;
          end;
          inc(i);
        end;
      end;
    until Text=EndPointer;
    Strings.Sorted:=false;
    Strings.CustomSort(SortByObject);
    result.Seek(0,soFromBeginning);
    Error:=false;
  finally
    if Error then begin
      result.Free;
      result:=nil;
    end;
  end;
end;

function GetLine(RC:TMemoryStream):integer;
var E:TElement;
begin
  if Assigned(RC) and (RC.Position+sizeof(E)<RC.Size) then begin
    Rc.Read(E,sizeof(E));
    result:=E.LineID shr 8;
  end
  else result:=-1;
end;

function TestStreamLine(Text:TMemoryStream;Pattern:string;var Values:array of Integer;Move:boolean):integer;
var i,n,Old:integer;
    Element:TElement;
    Error:boolean;
begin
  i:=1;
  result:=0;
  Error:=true;
  Old:=Text.Position;
  try
    while (result>=0) and (i<=length(Pattern)) do begin
      Text.Read(Element,sizeof(Element));
      if (Pattern[i]='|') then begin
        if ((Element.LineID and EL_SEPARATOR)=EL_SEPARATOR) and (Element.Info=2) then begin
          if not Move then Text.Seek(Old,soFromBeginning);
          Error:=false;
          exit;
        end
        else inc(i);
      end;
      case Pattern[i] of
        'K':n:=EL_KEYWORD;
        'T':n:=EL_TEXT;
        'C':n:=EL_CONSTANT;
        'N':n:=EL_NAME;
        ' ',',','#':n:=EL_SEPARATOR;
        '$':n:=EL_NAME_OR_CONST;
        '%':n:=EL_TEXT_OR_CONST;
        else n:=0;
      end;
      if (Element.LineID and n)=0 then result:=-1
      else begin
        if n=EL_KEYWORD then begin
          n:=i+1;
          while (n<=length(Pattern)) and (Pattern[n]>='0') and (Pattern[n]<='9') do inc(n);
          if Element.Info<>StrToInt(copy(Pattern,i+1,n-i-1)) then result:=-1;
          i:=n;
        end
        else if n=EL_SEPARATOR then begin
          case Pattern[i] of
            ',':if Element.Info<>0 then result:=-1;
            ' ':if Element.Info<>1 then result:=-1;
            '#':if Element.Info<>2 then result:=-1;
            else result:=-1;
          end;
          inc(i);
        end
        else begin
          if (Pattern[i]='$') or (Pattern[i]='%') then begin
            Values[result]:=Element.LineID and $FF;
            inc(result);
          end;
          Values[result]:=Element.Info;
          inc(result);
          inc(i);
        end;
      end;
    end;
    if result>=0 then begin
      Text.Read(Element,sizeof(Element));
      if (not TestFlag(Element.LineID,EL_SEPARATOR)) or (Element.Info<>2) then result:=-1;
    end;
    Error:=false;
  finally
    if Error then result:=-1;
  end;
  if not Move or (result=-1) then Text.Seek(Old,soFromBeginning);
end;

{****  RT_BITMAP  *************************************************************}

function Res2Bmp(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer):integer;
var BMPfh:PBitmapFileHeader;
    BMPih:PBitmapInfoHeader;
    ColorTableSize:cardinal;
begin
  result:=sizeof(TBitmapFileHeader)+SourceSize;
  if DestSize=CAN_CREATE then GetMem(Dest,result);
  if (DestSize>=result) or (DestSize=CAN_CREATE) then begin
    BMPfh:=Dest;
    ZeroMemory(BMPfh,sizeof(TBitmapFileHeader));
    BMPih:=ShiftPointer(Dest,sizeof(TBitmapFileHeader));
    Move(Source^,BMPih^,SourceSize);
    BMPfh^.bfType:=$4D42;
    BMPfh^.bfSize:=sizeof(TBitmapFileHeader)+SourceSize;
    if BMPih^.biClrUsed=0 then begin
      if BMPih^.biBitCount<=8 then ColorTableSize:=4*(1 shl BMPih^.biBitCount)
      else ColorTableSize:=0;
    end
    else ColorTableSize:=4*BMPih^.biClrUsed;
    BMPfh^.bfOffBits:=sizeof(TBitmapFileHeader)+BMPih^.biSize+ColorTableSize;
  end;
end;

function Bmp2Res(Source:pointer;SourceSize:integer;var Dest:pointer):integer;
var Info:pointer;
begin
  result:=SourceSize-sizeof(TBitmapFileHeader);
  GetMem(Dest,result);
  Info:=ShiftPointer(Source,sizeof(TBitmapFileHeader));
  Move(Info^,Dest^,result);
end;

{**** RT_MENU  ****************************************************************}

type TMenuHeader = packed record
       wVersion:word;
       cbHeaderSize:word;
     end;
     TMenuItem = packed record
       wResInfo:word;
       ID:word;
     end;
     TMenuExHeader = packed record
       wVersion:word;
       wOffset:word;
       dwHelp:cardinal;
     end;
     TMenuExItem = packed record
       dwType:cardinal;
       dwState:cardinal;
       uID:cardinal;
       wResInfo:word;
     end;

function IDToMenuFlag(ResInfo:word):string;
begin
  result:='';
  if TestFlag(ResInfo,MFS_GRAYED) then result:=result+', GRAYED';
  if TestFlag(ResInfo,MFS_CHECKED) then result:=result+', CHECKED';
  if TestFlag(ResInfo,MF_MENUBARBREAK) then result:=result+', MENUBARBREAK';
  if TestFlag(ResInfo,MF_MENUBREAK) then result:=result+', MENUBREAK';
  if TestFlag(ResInfo,MF_HELP) then result:=result+', HELP';
end;

function Res2Menu(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;ResName:PChar;ResLang:word):integer;
var Text:TStringList;
    Flags,Name:string;
    Level,EndPointer:NativeUInt;
    ACL:array[0..63] of boolean;
    EndFlag:boolean;
    ID,dwType,dwState,dwHelpID:cardinal;
begin
  if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
  else Text:=nil;
  result:=0;
  EndPointer:= NativeUInt(ShiftPointer(Source,SourceSize));
  try
    if (TMenuHeader(Source^).wVersion=0) and (TMenuHeader(Source^).cbHeaderSize=0) then begin
      Level:=1;
      result:=AddString(Text,result,NameToStr(ResName)+' MENU');
      result:=AddString(Text,result,LanguageString(ResLang));
      result:=AddString(Text,result,'BEGIN');
      Source:=ShiftPointer(Source,4);
      repeat;
        if TestFlag(TMenuItem(Source^).wresInfo,MF_POPUP) then begin
          ACL[Level+1]:=TestFlag(TMenuItem(Source^).wresInfo,MF_END);
          Flags:=IDToMenuFlag(TMenuItem(Source^).wresInfo and $FF6F);
          Source:=ShiftPointer(Source,2);
          Name:=WideCharToString(PWideChar(Source));
          result:=AddString(Text,result,StringOfChar(' ',Level*2)+'POPUP "'+StrToCStyle(Name)+'"'+Flags);
          result:=AddString(Text,result,StringOfChar(' ',Level*2)+'BEGIN');
          inc(Level);
          Source:=ShiftPointer(Source,2*length(Name)+2);
        end
        else begin
          EndFlag:=TestFlag(TMenuItem(Source^).wresInfo,MF_END);
          Flags:=IDToMenuFlag(TMenuItem(Source^).wresInfo and $FF6F);
          ID:=TMenuItem(Source^).ID;
          Source:=ShiftPointer(Source,4);
          if word(Source^)=0 then begin
            Name:='SEPARATOR';
            Source:=ShiftPointer(Source,2);
          end
          else begin
            Name:=WideCharToString(PWideChar(Source));
            Source:=ShiftPointer(Source,2*length(Name)+2);
            Name:='"'+StrToCStyle(Name)+'", '+IntToStr(ID);
          end;
          result:=AddString(Text,result,StringOfChar(' ',Level*2)+'MENUITEM '+Name+Flags);
          if EndFlag then begin
            repeat;
              dec(Level);
              result:=AddString(Text,result,StringOfChar(' ',Level*2)+'END');
            until (Level<=0) or (not ACL[Level+1]);
          end;
        end;
      until NativeUInt(Source)>=EndPointer;
    end
    else if TMenuExHeader(Source^).wVersion=1 then begin
      Level:=1;
      result:=AddString(Text,result,NameToStr(ResName)+' MENUEX');
      result:=AddString(Text,result,LanguageString(ResLang));
      result:=AddString(Text,result,'BEGIN');
      Source:=ShiftPointer(Source,4+TMenuExHeader(Source^).wOffset);
      repeat;
        if TestFlag(TMenuExItem(Source^).wResInfo,1) then begin
          ACL[Level+1]:=TestFlag(TMenuExItem(Source^).wResInfo,MF_END);
          dwType:=TMenuExItem(Source^).dwType;
          dwState:=TMenuExItem(Source^).dwState;
          ID:=TMenuExItem(Source^).uID;
          Source:=ShiftPointer(Source,14);
          Name:=WideCharToString(PWideChar(Source));
          Source:=ShiftPointer(Source,2*length(Name)+2);
          if (length(Name) and 1)=1 then Source:=ShiftPointer(Source,2);
          dwHelpID:=cardinal(Source^);
          Source:=ShiftPointer(Source,4);
          if dwHelpID<>0 then Flags:=', '+IntToStr(ID)+', 0x'+IntToHex(dwType,8)+', 0x'+IntToHex(dwState,8)+', 0x'+IntToHex(dwHelpID,8)
          else if dwState<>0 then Flags:=', '+IntToStr(ID)+', 0x'+IntToHex(dwType,8)+', 0x'+IntToHex(dwState,8)
          else if dwType<>0 then Flags:=', '+IntToStr(ID)+', 0x'+IntToHex(dwType,8)
          else if ID<>0 then Flags:=', '+IntToStr(ID)
          else Flags:='';
          result:=AddString(Text,result,StringOfChar(' ',Level*2)+'POPUP "'+StrToCStyle(Name)+'"'+Flags);
          result:=AddString(Text,result,StringOfChar(' ',Level*2)+'BEGIN');
          inc(Level);
        end
        else begin
          EndFlag:=TestFlag(TMenuExItem(Source^).wResInfo,MF_END);
          dwType:=TMenuExItem(Source^).dwType;
          dwState:=TMenuExItem(Source^).dwState;
          ID:=TMenuExItem(Source^).uID;
          Source:=ShiftPointer(Source,14);
          Name:=WideCharToString(PWideChar(Source));
          Source:=ShiftPointer(Source,2*length(Name)+2);
          if (length(Name) and 1)=1 then Source:=ShiftPointer(Source,2);
          if dwState<>0 then Flags:=', '+IntToStr(ID)+', 0x'+IntToHex(dwType,8)+', 0x'+IntToHex(dwState,8)
          else if dwType<>0 then Flags:=', '+IntToStr(ID)+', 0x'+IntToHex(dwType,8)
          else if ID<>0 then Flags:=', '+IntToStr(ID)
          else Flags:='';
          if TestFlag(dwType,MFT_SEPARATOR) then result:=AddString(Text,result,StringOfChar(' ',Level*2)+'MENUITEM SEPARATOR')
          else result:=AddString(Text,result,StringOfChar(' ',Level*2)+'MENUITEM "'+StrToCStyle(Name)+'"'+Flags);
          if EndFlag then begin
            repeat;
              dec(Level);
              result:=AddString(Text,result,StringOfChar(' ',Level*2)+'END');
            until (Level<=0) or (not ACL[Level+1]);
          end;
        end;
      until NativeUInt(Source)>=EndPointer;
    end
    else result:=UNKNOWN_FORMAT;
  finally
    if Assigned(Text) then begin
      if ((DestSize=CAN_CREATE) or (DestSize>=result)) and (result<>UNKNOWN_FORMAT) then begin
        Result := Result * 2 + 2;
        if Text.Count > 0 then
          Text[0] := Char($FEFF) + Text[0];
        if DestSize=CAN_CREATE then GetMem(Dest,result);
        Move(PChar(Text.Text)^,Dest^,result);
      end;
      Text.Free;
    end;
  end;
end;

function Menu2Res(Source:pointer;SourceSize:integer;var Dest:pointer;ResName:PChar;ResLang:word):integer;
var RC:TMemoryStream;
    OutPut:TMemoryStream;
    Strings:TStringList;
    Ints:array[0..7] of integer;
    Element:TElement;
    i,n,Level:integer;
    Flags:word;
    WideStr:array[0..2047] of word;
    PWideStr:PWideChar;
    MenuItem:TMenuExItem;
begin
  RC:=nil; PWideStr:=PWideChar(@WideStr);
  result:=UNKNOWN_FORMAT;
  Strings:=TStringList.Create;
  OutPut:=TMemoryStream.Create;
  try
    RC:=LoadResource(Source,SourceSize,Strings);
    if Assigned(RC) then begin
      if TestStreamLine(RC,'$ K15',Ints,true)=2 then begin
        if TestStreamLine(RC,'K3 C',Ints,true)<>1 then TestStreamLine(RC,'K3 C,C',Ints,true);
        if TestStreamLine(RC,'K11',Ints,true)=0 then begin
          i:=0; OutPut.Write(i,4); Level:=1;
          repeat
            ZeroMemory(@Ints,sizeof(Ints));
            if TestStreamLine(RC,'K10 T|,C',Ints,true)>=0 then begin
              if TestStreamLine(RC,'K11',Ints,true)=-1 then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': "BEGIN" expected');
              inc(Level);
              i:=RC.Position; n:=1;
              repeat;
                RC.Read(Element,sizeof(Element));
                if TestFlag(Element.LineID,EL_KEYWORD) then begin
                  if Element.Info=11 then inc(n)
                  else if Element.Info=16 then dec(n);
                end;
              until n=0;
              RC.Read(Element,sizeof(Element));
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (Element.Info=16) then Flags:=MF_END else Flags:=0;
              RC.Seek(i,soFromBeginning);
              Flags:=Flags or Ints[1] or MF_POPUP;
              StringToWideChar(Strings[Ints[0]],PWideStr,2048);
              OutPut.Write(Flags,2);
              OutPut.Write(WideStr,2*length(Strings[Ints[0]])+2);
            end
            else if TestStreamLine(RC,'K2 T,C|,C',Ints,true)>=0 then begin
              i:=RC.Position;
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (element.Info=16) then Flags:=MF_END else Flags:=0;
              RC.Seek(i,soFromBeginning);
              Flags:=Flags or Ints[2];
              StringToWideChar(Strings[Ints[0]],PWideStr,2048);
              OutPut.Write(Flags,2);
              OutPut.Write(Ints[1],2);
              OutPut.Write(WideStr,2*length(Strings[Ints[0]])+2);
            end
            else if TestStreamLine(RC,'K2 K1',Ints,true)>=0 then begin
              i:=RC.Position; n:=0;
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (element.Info=16) then Flags:=MF_END else Flags:=0;
              RC.Seek(i,soFromBeginning);
              OutPut.Write(Flags,2);
              OutPut.Write(n,4);
            end
            else if TestStreamLine(RC,'K16',Ints,true)>=0 then dec(Level)
            else raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknown statement');
          until Level=0;
        end;
      end
      else if TestStreamLine(RC,'$ K9',Ints,true)=2 then begin
        if TestStreamLine(RC,'K3 C',Ints,true)<>1 then TestStreamLine(RC,'K3 C,C',Ints,true);
        if TestStreamLine(RC,'K11',Ints,true)=0 then begin
          i:=$00040001; OutPut.Write(i,4); i:=0; OutPut.Write(i,4);
          Level:=1;
          repeat
            ZeroMemory(@Ints,sizeof(Ints));
            if TestStreamLine(RC,'K10 T|,C|,C|,C|,C',Ints,true)>=0 then begin
              if TestStreamLine(RC,'K11',Ints,true)=-1 then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': "BEGIN" expected');
              inc(Level);
              i:=RC.Position; n:=1;
              repeat;
                RC.Read(Element,sizeof(Element));
                if TestFlag(Element.LineID,EL_KEYWORD) then begin
                  if Element.Info=11 then inc(n)
                  else if Element.Info=16 then dec(n);
                end;
              until n=0;
              RC.Read(Element,sizeof(Element));
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (Element.Info=16) then Flags:=MF_END else Flags:=0;
              RC.Seek(i,soFromBeginning);
              MenuItem.dwType:=Ints[2];
              MenuItem.dwState:=Ints[3];
              MenuItem.uID:=Ints[1];
              MenuItem.wResInfo:=Flags or 1;
              OutPut.Write(MenuItem,sizeof(MenuItem));
              StringToWideChar(Strings[Ints[0]],PWideStr,2048);
              i:=length(Strings[Ints[0]]);
              OutPut.Write(WideStr,2*i+2);
              if (i and 1)=1 then begin
                i:=0; OutPut.Write(i,2);
              end;
              OutPut.Write(Ints[4],4);
            end
            else if TestStreamLine(RC,'K2 T,C|,C|,C',Ints,true)>=0 then begin
              i:=RC.Position;
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (element.Info=16) then MenuItem.wResInfo:=MF_END else MenuItem.wResInfo:=0;
              RC.Seek(i,soFromBeginning);
              MenuItem.dwType:=Ints[2];
              MenuItem.dwState:=Ints[3];
              MenuItem.uID:=Ints[1];
              OutPut.Write(MenuItem,sizeof(MenuItem));
              StringToWideChar(Strings[Ints[0]],PWideStr,2048);
              i:=length(Strings[Ints[0]]);
              OutPut.Write(WideStr,2*i+2);
              if (i and 1)=1 then begin
                i:=0; OutPut.Write(i,2);
              end;
            end
            else if TestStreamLine(RC,'K2 K1',Ints,true)>=0 then begin
              i:=RC.Position;
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_KEYWORD) and (element.Info=16) then MenuItem.wResInfo:=MF_END else MenuItem.wResInfo:=0;
              RC.Seek(i,soFromBeginning);
              MenuItem.dwType:=MFT_SEPARATOR;
              MenuItem.dwState:=0;
              MenuItem.uID:=0;
              OutPut.Write(MenuItem,sizeof(MenuItem));
              i:=0; OutPut.Write(i,2);
            end
            else if TestStreamLine(RC,'K16',Ints,true)>=0 then dec(Level)
            else raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknown statement');
          until Level=0;
        end;
      end;
      GetMem(Dest,OutPut.Size);
      Move(OutPut.Memory^,Dest^,OutPut.Size);
      result:=OutPut.Size;
    end;
  finally
    Strings.Free;
    OutPut.Free;
    if Assigned(RC) then RC.Free;
  end;
end;

{****  RT_DIALOG  *************************************************************}

type TDialogHeader = packed record // Size = 18 bytes
       dwStyle:cardinal;
       dwExtStyle:cardinal;
       wControl:word;
       x,y,cx,cy:word;
     end;
     TDialogItem = packed record
       dwStyle:cardinal;
       dwExtStyle:cardinal;
       x,y,cx,cy:word;
       ID:smallint;
     end;
     TDialogExHeader = packed record
       wVersion:word;
       wSign:word;
       dwHelpID:cardinal;
       dwExtStyle:cardinal;
       dwStyle:cardinal;
       wControl:word;
       x,y,cx,cy:word;
     end;
     TDialogExItem = packed record
       dwHelpID:cardinal;
       dwExtStyle:cardinal;
       dwStyle:cardinal;
       x,y,cx,cy:word;
       ID:integer;
     end;

function IDToDialogItem(ID:word):string;
begin
  case ID of
    $0080:result:='BUTTON';
    $0081:result:='EDIT';
    $0082:result:='STATIC';
    $0084:result:='SCROLLBAR';
    $0085:result:='COMBOBOX';
    else result:=IntToStr(ID);
  end;
end;

function Res2Dlg(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;ResName:PChar;ResLang:word):integer;
var Text:TStringList;
    Count,Shift,i,n:integer;
    Extended,SetFont,Error:boolean;
    DlgItem:TDialogItem;
    DlgItemEx:TDialogExItem;
    s,CtrlClass,CtrlName:string;
begin
  if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
  else Text:=nil;
  result:=0;
  Error:=true;
  try
    if cardinal(Source^)=$FFFF0001 then begin
      with TDialogExHeader(Source^) do begin
          if dwHelpID<>0 then s:=', 0x'+IntToHex(dwHelpID,8) else s:='';
          result:=AddString(Text,0,NameToStr(ResName)+' DIALOGEX '+IntToStr(x)+', '+IntToStr(y)+', '+IntToStr(cx)+', '+IntToStr(cy)+s);
          result:=AddString(Text,result,'STYLE 0x'+IntToHex(dwStyle,8));
          if dwExtStyle<>0 then result:=AddString(Text,result,'EXSTYLE 0x'+IntToHex(dwExtStyle,8));
          SetFont:=TestFlag(dwStyle,DS_SETFONT);
          Count:=wControl;
      end;
      Extended:=true; Shift:=13;
      Source:=ShiftPointer(Source,26);
    end
    else begin
      with TDialogHeader(Source^) do begin
          result:=AddString(Text,result,NameToStr(ResName)+' DIALOG '+IntToStr(x)+', '+IntToStr(y)+', '+IntToStr(cx)+', '+IntToStr(cy));
          result:=AddString(Text,result,'STYLE 0x'+IntToHex(dwStyle,8));
          if dwExtStyle<>0 then result:=AddString(Text,result,'EXSTYLE 0x'+IntToHex(dwExtStyle,8));
          SetFont:=TestFlag(dwStyle,DS_SETFONT);
          Count:=wControl;
      end;
      Extended:=false; Shift:=9;
      Source:=ShiftPointer(Source,18);
    end;
    if word(Source^)<>0 then begin
      if word(Source^)=$FFFF then begin
          Source:=ShiftPointer(Source,2); inc(Shift);
          result:=AddString(Text,result,'MENU '+IntToStr(word(Source^)));
      end
      else begin
          s:=WideCharToString(PWideChar(Source));
          result:=AddString(Text,result,'MENU "'+StrToCStyle(s)+'"');
          Source:=ShiftPointer(Source,2*length(s)); Shift:=Shift+length(s);
      end;
    end;
    Source:=ShiftPointer(Source,2); inc(Shift);
    if word(Source^)<>0 then begin
      if word(Source^)=$FFFF then begin
          Source:=ShiftPointer(Source,2); inc(Shift);
          result:=AddString(Text,result,'CLASS '+IntToStr(word(Source^)));
      end
      else begin
          s:=WideCharToString(PWideChar(Source));
          result:=AddString(Text,result,'CLASS "'+StrToCStyle(s)+'"');
          Source:=ShiftPointer(Source,2*length(s)); Shift:=Shift+length(s);
      end;
    end;
    Source:=ShiftPointer(Source,2); inc(Shift);
    if word(Source^)<>0 then begin
      s:=WideCharToString(PWideChar(Source));
      result:=AddString(Text,result,'CAPTION "'+StrToCStyle(s)+'"');
      Source:=ShiftPointer(Source,2*length(s)); Shift:=Shift+length(s);
    end;
    Source:=ShiftPointer(Source,2); inc(Shift);
    if SetFont then begin
      i:=word(Source^);
      if Extended then begin
          Source:=ShiftPointer(Source,6); Shift:=Shift+3;
      end
      else begin
          Source:=ShiftPointer(Source,2); inc(Shift);
      end;
      s:=WideCharToString(PWideChar(Source));
      result:=AddString(Text,result,'FONT '+IntToStr(i)+', "'+StrToCStyle(s)+'"');
      Source:=ShiftPointer(Source,2*length(s)+2); Shift:=Shift+length(s)+1;
    end;
    result:=AddString(Text,result,LanguageString(ResLang));
    result:=AddString(Text,result,'BEGIN');
    while Count>0 do begin
      if (Shift and 1)<>0 then Source:=ShiftPointer(Source,2);
      if Extended then begin
          DlgItemEx:=TDialogExItem(Source^); Shift:=12;
      end
      else begin
          DlgItem:=TDialogItem(Source^); Shift:=9;
      end;
      Source:=ShiftPointer(Source,Shift*2);
      if word(Source^)=$FFFF then begin
          Source:=ShiftPointer(Source,2);
          CtrlClass:=IDToDialogItem(word(Source^));
          Source:=ShiftPointer(Source,2);
          Shift:=Shift+2;
      end
      else begin
          CtrlClass:=WideCharToString(PWideChar(Source));
          Source:=ShiftPointer(Source,2*length(CtrlClass)+2);
          Shift:=Shift+length(CtrlClass)+1;
          CtrlClass:='"'+StrToCStyle(CtrlClass)+'"';
      end;
      if word(Source^)=$FFFF then begin
          Source:=ShiftPointer(Source,2);
          CtrlName:=IntToStr(word(Source^));
          Source:=ShiftPointer(Source,2);
          Shift:=Shift+2;
      end
      else begin
          CtrlName:=WideCharToString(PWideChar(Source));
          Source:=ShiftPointer(Source,2*length(CtrlName)+2);
          Shift:=Shift+length(CtrlName)+1;
          CtrlName:='"'+StrToCStyle(CtrlName)+'"';
      end;
      if Extended then begin
          s:=', 0x'+IntToHex(DlgItemEx.dwStyle,8)+', '+IntToStr(DlgItemEx.x)+', '+IntToStr(DlgItemEx.y)+', '+IntToStr(DlgItemEx.cx)+', '+IntToStr(DlgItemEx.cy);
          if (DlgItemEx.dwExtStyle<>0) or (DlgItemEx.dwhelpID<>0) then s:=s+', 0x'+IntToHex(DlgItemEx.dwExtStyle,8)+', 0x'+IntToHex(DlgItemEx.dwhelpID,8);
          result:=AddString(Text,result,'  CONTROL '+CtrlName+', '+IntToStr(DlgItemEx.ID)+', '+CtrlClass+s);
      end
      else begin
          s:=', 0x'+IntToHex(DlgItem.dwStyle,8)+', '+IntToStr(DlgItem.x)+', '+IntToStr(DlgItem.y)+', '+IntToStr(DlgItem.cx)+', '+IntToStr(DlgItem.cy);
          if DlgItem.dwExtStyle<>0 then s:=s+', 0x'+IntToHex(DlgItem.dwExtStyle,8);
          result:=AddString(Text,result,'  CONTROL '+CtrlName+', '+IntToStr(DlgItem.ID)+', '+CtrlClass+s);
      end;
      if word(Source^)>0 then begin
          result:=AddString(Text,result,'  BEGIN');
          i:=word(Source^); n:=2; s:='';
          while i>0 do begin
            s:=s+IntToStr(TByteArray(Source^)[n]);
            if i>1 then s:=s+',';
            if ((n-1) mod 10=0) or (i=1) then begin
                result:=AddString(Text,result,'      '+s);
                s:='';
            end
            else s:=s+' ';
            dec(i); inc(n);
          end;
          result:=AddString(Text,result,'  END');
      end;
      i:=(word(Source^)+1) div 2+1;
      Source:=ShiftPointer(Source,2*i); Shift:=Shift+i;
      dec(Count);
    end;
    result:=AddString(TEXT,result,'END');
    Error:=false;
  finally
    if Error then result:=UNKNOWN_FORMAT;
    if Assigned(Text) then begin
      if ((DestSize=CAN_CREATE) or (DestSize>=result)) and (result<>UNKNOWN_FORMAT) then begin
        Result := Result * 2 + 2;
        if Text.Count > 0 then
          Text[0] := Char($FEFF) + Text[0];
        if DestSize=CAN_CREATE then GetMem(Dest,result);
        Move(PChar(Text.Text)^,Dest^,result);
      end;
      Text.Free;
    end;
  end;
end;

function Dlg2Res(Source:pointer;SourceSize:integer;var Dest:pointer;ResName:PChar;ResLang:word):integer;
var RC:TMemoryStream;
    OutPut:TMemoryStream;
    Strings:TStringList;
    Ints:array[0..31] of integer;
    Extended,SetFont,Dialog:boolean;
    DlgItem:TDialogItem;
    DlgItemEx:TDialogExItem;
    DlgHeader:TDialogHeader;
    DlgHeaderEx:TDialogExHeader;
    DlgCaption,DlgClass,DlgFont,DlgMenu:string;
    DlgFontSize:word;
    Zero,FFFF,i:integer;
    WideStr:array[0..2047] of word;
    PWideStr:PWideChar;
    Buffer:array[0..65535] of byte;
    Element:TElement;
begin
  RC:=nil; PWideStr:=PWideChar(@WideStr);
  result:=UNKNOWN_FORMAT;
  ZeroMemory(@DlgHeader,sizeof(DlgHeader));
  ZeroMemory(@DlgHeaderEx,sizeof(DlgHeaderEx));
  Strings:=TStringList.Create;
  OutPut:=TMemoryStream.Create;
  Extended:=false; Dialog:=true;
  try
    RC:=LoadResource(Source,SourceSize,Strings);
    if Assigned(RC) then begin
      ZeroMemory(@Ints,sizeof(Ints));
      if TestStreamLine(RC,'$ K8 C,C,C,C',Ints,true)=6 then begin
        with DlgHeader do begin
          x:=Ints[2];
          y:=Ints[3];
          cx:=Ints[4];
          cy:=Ints[5];
          wControl:=0;
        end;
      end
      else if TestStreamLine(RC,'$ K4 C,C,C,C|,C',Ints,true)>5 then begin
        Extended:=true;
        with DlgHeaderEx do begin
          x:=Ints[2];
          y:=Ints[3];
          cx:=Ints[4];
          cy:=Ints[5];
          wVersion:=1;
          wSign:=$FFFF;
          dwHelpID:=Ints[6];
          wControl:=0;
        end;
      end
      else Dialog:=false;
      if Dialog then begin
        DlgCaption:=''; DlgMenu:=''; DlgClass:=''; DlgFont:=''; DlgFontSize:=0;
        Zero:=0; FFFF:=integer($FFFFFFFF); SetFont:=false;
        repeat;
          if TestStreamLine(RC,'K6 T',Ints,true)=1 then DlgCaption:=Strings[Ints[0]]
          else if TestStreamLine(RC,'K12 C',Ints,true)=1 then DlgClass:=IntToStr(Ints[0])
          else if TestStreamLine(RC,'K12 T',Ints,true)=1 then DlgClass:=Strings[Ints[0]]
          else if TestStreamLine(RC,'K13 C',Ints,true)=1 then begin
            DlgHeader.dwStyle:=Ints[0];
            DlgHeaderEx.dwStyle:=Ints[0];
            SetFont:=TestFlag(Ints[0],DS_SETFONT);
          end
          else if TestStreamLine(RC,'K15 C',Ints,true)=1 then DlgMenu:=IntToStr(Ints[0])
          else if TestStreamLine(RC,'K15 T',Ints,true)=1 then DlgMenu:=Strings[Ints[0]]
          else if TestStreamLine(RC,'K5 C',Ints,true)=1 then begin
            DlgHeader.dwExtStyle:=Ints[0];
            DlgHeaderEx.dwExtStyle:=Ints[0];
          end
          else if TestStreamLine(RC,'K14 C,T',Ints,true)=2 then begin
            DlgFontSize:=Ints[0];
            DlgFont:=Strings[Ints[1]];
          end
          else if (TestStreamLine(RC,'K11',Ints,true)<>0) and (TestStreamLine(RC,'K3 C',Ints,true)<>1) and (TestStreamLine(RC,'K3 C,C',Ints,true)<>2) then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknows statement');
        until TestStreamLine(RC,'K11',Ints,true)=0;
        if Extended then OutPut.Write(DlgHeaderEx,sizeof(DlgHeaderEx)) else OutPut.Write(DlgHeader,sizeof(DlgHeader));
        if DlgMenu<>'' then begin
          if IsDec(DlgMenu) then begin
            OutPut.Write(FFFF,2); i:=StrToInt(DlgMenu); OutPut.Write(i,2);
          end
          else begin
            StringToWideChar(DlgMenu,PWideStr,2048);
            OutPut.Write(WideStr,2*length(DlgMenu)+2);
          end;
        end
        else OutPut.Write(Zero,2);
        if DlgClass<>'' then begin
          if IsDec(DlgClass) then begin
            OutPut.Write(FFFF,2); i:=StrToInt(DlgClass); OutPut.Write(i,2);
          end
          else begin
            StringToWideChar(DlgClass,PWideStr,2048);
            OutPut.Write(WideStr,2*length(DlgClass)+2);
          end;
        end
        else OutPut.Write(Zero,2);
        if DlgCaption<>'' then begin
          StringToWideChar(DlgCaption,PWideStr,2048);
          OutPut.Write(WideStr,2*length(DlgCaption)+2);
        end
        else OutPut.Write(Zero,2);
        if SetFont and (DlgFont<>'') and (DlgFontSize<>0) then begin
          OutPut.Write(DlgFontSize,2);
          if Extended then begin
            i:=$01000000; OutPut.Write(i,4);
          end;
          StringToWideChar(DlgFont,PWideStr,2048);
          OutPut.Write(WideStr,2*length(DlgFont)+2);
        end;
        while TestStreamLine(RC,'K16',Ints,true)<>0 do begin
          ZeroMemory(@Ints,sizeof(Ints));
          i:=OutPut.Size mod 4;
          if i>0 then OutPut.Write(Zero,4-i);
          if Extended then begin
            i:=TestStreamLine(RC,'K7 %,C,%,C,C,C,C,C|,C|,C',Ints,true);
            if (i<10) or (i>12) then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknown statement');
            with DlgItemEx do begin
              dwStyle:=Ints[5];
              dwExtStyle:=Ints[10];
              dwHelpID:=Ints[11];
              x:=Ints[6];
              y:=Ints[7];
              cx:=ints[8];
              cy:=ints[9];
              ID:=Ints[2];
            end;
            OutPut.Write(DlgItemEx,sizeof(DlgItemEx));
          end
          else begin
            i:=TestStreamLine(RC,'K7 %,C,%,C,C,C,C,C|,C',Ints,true);
            if (i<10) or (i>11) then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknown statement');
            with DlgItem do begin
              dwStyle:=Ints[5];
              dwExtStyle:=Ints[10];
              x:=Ints[6];
              y:=Ints[7];
              cx:=Ints[8];
              cy:=Ints[9];
              ID:=Ints[2];
            end;
            OutPut.Write(DlgItem,sizeof(DlgItem));
          end;
          if Ints[3]=EL_CONSTANT then begin
            OutPut.Write(FFFF,2); OutPut.Write(Ints[4],2);
          end
          else begin
            StringToWideChar(Strings[Ints[4]],PWideStr,2048);
            OutPut.Write(WideStr,2*length(Strings[Ints[4]])+2);
          end;
          if Ints[0]=EL_CONSTANT then begin
            OutPut.Write(FFFF,2); OutPut.Write(Ints[1],2);
          end
          else begin
            StringToWideChar(Strings[Ints[1]],PWideStr,2048);
            OutPut.Write(WideStr,2*length(Strings[Ints[1]])+2);
          end;
          if TestStreamLine(RC,'K11',Ints,true)=0 then begin
            i:=0;
            repeat;
              RC.Read(Element,sizeof(Element));
              if TestFlag(Element.LineID,EL_CONSTANT) then begin
                Buffer[i]:=Element.Info;
                inc(i);
                RC.Read(Element,sizeof(Element));
                if not TestFlag(Element.LineID,EL_SEPARATOR) then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Separator (space, comma or new line) expected');
              end
              else raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Integer value expected');
            until TestStreamLine(RC,'K16',Ints,true)=0;
            OutPut.Write(i,2);
            OutPut.Write(Buffer,i);
          end
          else OutPut.Write(Zero,2);
          inc(DlgHeader.wControl);
          inc(DlgHeaderEx.wControl);
        end;
        GetMem(Dest,OutPut.Size);
        Move(OutPut.Memory^,Dest^,OutPut.Size);
        if Extended then Move(DlgHeaderEx,Dest^,sizeof(DlgHeaderEx)) else Move(DlgHEader,Dest^,sizeof(DlgHeader));
        result:=OutPut.Size;
      end;
    end;
  finally
    if Assigned(RC) then RC.Free;
    Strings.Free;
    OutPut.Free;
  end;
end;

{****  RT_STRING  *************************************************************}

function Res2Str(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;ResName:PChar;ResLang:word):integer;
var Text:TStringList;
    i,n,v:integer;
    s:string;
    EndPointer: NativeUInt;
begin
  if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
  else Text:=nil;
  result:=0;
  try
    EndPointer:=NativeUInt(ShiftPointer(Source,SourceSize));
    result:=AddString(Text,0,'STRINGTABLE');
    result:=AddString(Text,result,LanguageString(ResLang));
    result:=AddString(Text,result,'BEGIN');
    n:=integer(ResName)*16-16; v:=0;
    repeat;
      if word(Source^)<>0 then begin
        i:=word(Source^); Source:=ShiftPointer(Source,2);
        WideCharLenToStrVar(PWideChar(Source),i,s);
        Source:=ShiftPointer(Source,2*i);
        result:=AddString(Text,result,'  '+IntToStr(n+v)+', "'+StrToCStyle(s)+'"');
      end
      else Source:=ShiftPointer(Source,2);
      inc(v);
    until (v>15) or (NativeUInt(Source)>=EndPointer);
    result:=AddString(Text,result,'END');
  finally
    if Assigned(Text) then begin
      if ((DestSize=CAN_CREATE) or (DestSize>=result)) and (result<>UNKNOWN_FORMAT) then begin
        Result := Result * 2 + 2;
        if Text.Count > 0 then
          Text[0] := Char($FEFF) + Text[0];
        if DestSize=CAN_CREATE then GetMem(Dest,result);
        Move(PChar(Text.Text)^,Dest^,result);
      end;
      Text.Free;
    end;
  end;
end;

function Str2Res(Source:pointer;SourceSize:integer;var Dest:pointer;ResName:PChar;ResLang:word):integer;
var Strings:TStringList;
    RC,Output:TMemoryStream;
    i,Size,First,Last:integer;
    s:array[0..15] of string;
    Ints:array[0..3] of integer;
    WideStr:array[0..2047] of word;
    PWideStr:PWideChar;
begin
  RC:=nil; PWideStr:=PWideChar(@WideStr);
  result:=UNKNOWN_FORMAT;
  Strings:=TStringList.Create;
  OutPut:=TMemoryStream.Create;
  try
    RC:=LoadResource(Source,SourceSize,Strings);
    if Assigned(RC) then begin
      if TestStreamLine(RC,'K0',Ints,true)>=0 then begin
        if TestStreamLine(RC,'K3 C',Ints,true)<>1 then TestStreamLine(RC,'K3 C,C',Ints,true);
        if TestStreamLine(RC,'K11',Ints,true)=0 then begin
          for i:=0 to 15 do s[i]:='';
          repeat;
            First:=integer(ResName)*16-16;
            Last:=First+15;
            if (TestStreamLine(RC,'C T',Ints,true)=2) or (TestStreamLine(RC,'C,T',Ints,true)=2) then begin
              if (Ints[0]<First) or (Ints[0]>Last) then raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Index '+IntToStr(Ints[0])+' cannot be in this string table.');
              s[Ints[0]-First]:=Strings[Ints[1]];
            end
            else raise EConvertError.Create('Error at line '+IntToStr(GetLine(RC))+': Unknown statement');
          until TestStreamLine(RC,'K16',Ints,true)>=0;
          for i:=0 to 15 do begin
            Size:=length(s[i]);
            OutPut.Write(Size,2);
            if Size>0 then begin
              StringToWideChar(s[i],PWideStr,2048);
              OutPut.Write(WideStr,2*Size);
            end;
          end;
        end;
      end;
      GetMem(Dest,OutPut.Size);
      Move(OutPut.Memory^,Dest^,OutPut.Size);
      result:=OutPut.Size;
    end;
  finally
    Strings.Free;
    OutPut.Free;
    if Assigned(RC) then RC.Free;
  end;
end;

{****  RT_CURSOR and RT_CURSORGROUP  ******************************************}

type TCURFileHeader = packed record
       idReserved:word;         //$0000
       idType:word;             //$0002
       idCount:word;            //$0001
       bWidth:byte;
       bHeight:byte;
       bColorCount:byte;        //$00 ???
       bReserved:byte;          //$00
       xHotSpot:word;
       yHotSpot:word;
       dwBytesInRes:cardinal;   //size of resource
       dwImageOffset:cardinal;  //$00000016
     end;
     PCURFileHeader = ^TCURFileHeader;

type TRawCURData = packed record
       xHotSpot:word;
       yHotSpot:word;
     end;
     PRawCURData = ^TRawCURData;

type TCursorGroup = packed record
       idReserved:word;
       idType:word;
       idCount:word;
     end;
     PCursorGroup = ^TCursorGroup;

type TCursorGroupEntry = packed record
       wWidth:word;
       wHeight:word;
       wPalnes:word;
       wBitCount:word;
       dwBytesInRes:cardinal;
       nID:word;
     end;
     TCursors = array[0..(MAXINT div sizeof(TCursorGroupEntry))-1] of TCursorGroupEntry;
     PCursors = ^TCursors;

function Cur2Res(Source:pointer;SourceSize:integer;var Dest:pointer):integer;
var Info1,Info2:pointer;
    RawData:PRawCURData;
    CURfh:PCURFileHeader;
begin
  result:=SourceSize-sizeof(TCURFileHeader)+sizeof(TRawCURData);
  GetMem(Dest,result);
  RawData:=Dest;
  CURfh:=Source;
  RawData^.xHotSpot:=CURfh^.xHotSpot;
  RawData^.yHotSpot:=CURfh^.yHotSpot;
  Info1:=ShiftPointer(Dest,sizeof(TRawCURData));
  Info2:=ShiftPointer(Source,sizeof(TCURFileHeader));
  Move(Info2^,Info1^,SourceSize-sizeof(TCURFileHeader));
end;

function Res2Cur(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer):integer;
var CURfh:PCURFileHeader;
    BMPih:PBitmapInfoHeader;
    RawData:PRawCURData;
begin
  result:=sizeof(TCURFileHeader)-sizeof(TRawCURData)+SourceSize;
  if DestSize=CAN_CREATE then GetMem(Dest,result);
  if (DestSize>=result) or (DestSize=CAN_CREATE) then begin
    CURfh:=Dest;
    BMPih:=ShiftPointer(Dest,sizeof(TCURFileHeader));
    RawData:=Source;
    Source:=ShiftPointer(Source,sizeof(TRawCURData));
    Move(Source^,BMPih^,SourceSize-sizeof(TRawCURData));
    with CURfh^ do begin
      bColorCount:=0;
      bHeight:=BMPih^.biHeight div 2;
      bReserved:=0;
      bWidth:=BMPih^.biWidth;
      dwBytesInRes:=SourceSize-4;
      dwImageOffset:=$16;
      idCount:=1;
      idReserved:=0;
      idType:=2;
      xHotSpot:=RawData^.xHotSpot;
      yHotSpot:=RawData^.yHotSpot;
    end;
  end;
end;

function Res2CursorGroup(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;ResName:PChar;ResLang:word):integer;
var Text:TStringList;
    Cursors:PCursors;
    Header:PCursorGroup;
    i:integer;
begin
  if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
  else Text:=nil;
  result:=0;
  try
    Header:=Source;
    Cursors:=ShiftPointer(Source,sizeof(TCursorGroup));
    result:=AddString(Text,result,'<html><head>');
    result:=AddString(Text,result,'<title>RT_GROUP_CURSOR :: '+NameToStr(ResName)+' :: '+IntToStr(ResLang)+'</title>');
    result:=AddString(Text,result,'<style type=text/css>');
    result:=AddString(Text,result,'body {margin: 0; font-family: helvetica, sans-serif; font-size: 80%}');
    result:=AddString(Text,result,'td {padding-left: 1 em; padding-right: 1 em; font-size: 80%}');
    result:=AddString(Text,result,'tr {background-color: #EEEEEE; padding-left: 1 em}');
    result:=AddString(Text,result,'h1 {margin: 1 em 2 em 0 em 2 em; padding: 0; font-size: 125%; text-transform: uppercase}');
    result:=AddString(Text,result,'p {margin: 1 em; text-align: justify;}');
    result:=AddString(Text,result,'.head {background-color: #666666; color: #FFFFFF; font-weight: bold}');
    result:=AddString(Text,result,'</style></head><body>');
    result:=AddString(Text,result,'<h1>Resource Information</h1>');
    result:=AddString(Text,result,'<p><table><tr><td width=50%>Resource type</td><td>RT_GROUP_ICON</td></tr>');
    result:=AddString(Text,result,'<tr><td width=50%>Resource name</td><td>'+NameToStr(ResName)+'</td></tr>');
    result:=AddString(Text,result,'<tr><td width=50%>Language</td><td>'+IntToStr(ResLang)+'</td></tr></table></p>');
    result:=AddString(Text,result,'<h1>Resource Content</h1>');
    result:=AddString(Text,result,'<p><table><tr class=head><td>#</td><td>Size</td><td>Bits Per Pixel</td><td>IconID</td></tr>');
    for i:=0 to Header^.idCount-1 do result:=AddString(Text,result,'<tr><td>'+IntToStr(i+1)+'</td><td>'+IntToStr(Cursors^[i].wWidth)+'x'+IntToStr(Cursors^[i].wHeight)+'</td><td>'+IntToStr(Cursors^[i].wBitCount)+' ('+IntToStr(1 shl Cursors^[i].wBitCount)+' colors)</td><td>'+IntToStr(Cursors^[i].nID)+'</td></tr>');
    result:=AddString(Text,result,'</table></p>');
    result:=AddString(Text,result,'<p><font color=red>NOTE: if you want modify this resource you must use raw binary data not this HTML file.</font></p>');
    result:=AddString(Text,result,'</body><html>');
  finally
    if Assigned(Text) then begin
      if ((DestSize=CAN_CREATE) or (DestSize>=result)) and (result<>UNKNOWN_FORMAT) then begin
        if DestSize=CAN_CREATE then GetMem(Dest,result);
        Move(Pointer(AnsiString(Text.Text))^,Dest^,result);
      end;
      Text.Free;
    end;
  end;
end;

{****  RT_ICON and  RT_ICONGROUP  *********************************************}

type TICOFileHeader = packed record
       idReserved:word;         //$0000
       idType:word;             //$0001
       idCount:word;            //$0001
       bWidth:byte;
       bHeight:byte;
       bColorCount:byte;        //if BitPerPixe >= 8 then $0000
       bReserved:byte;          //$0000
       wPlanes:word;
       wBitCount:word;
       dwBytesInRes:cardinal;   //size of resource
       dwImageOffset:cardinal;  //$00000016
     end;
     PICOFileHeader = ^TICOFileHeader;

type TIconGroup = TCursorGroup;
     PIconGroup = ^TIconGroup;

type TIconGroupEntry = packed record
       bWidth:byte;
       bHeight:byte;
       bColorCount:byte;
       bReserved:byte;
       wPlanes:word;
       wBitCount:word;
       dwBytesInRes:cardinal;
       nID:word;
     end;
     TIcons = array[0..(MAXINT div sizeof(TIconGroupEntry))-1] of TIconGroupEntry;
     PIcons = ^TIcons;

function Ico2Res(Source:pointer;SourceSize:integer;var Dest:pointer):integer;
var Info:pointer;
begin
  result:=SourceSize-sizeof(TICOFileHeader);
  GetMem(Dest,result);
  Info:=ShiftPointer(Source,sizeof(TICOFileHeader));
  Move(Info^,Dest^,result);
end;

function Res2Ico(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer):integer;
var ICOfh:PICOFileHeader;
    BMPih:PBitmapInfoHeader;
begin
  result:=sizeof(TICOFileHeader)+SourceSize;
  if DestSize=CAN_CREATE then GetMem(Dest,result);
  if (DestSize>=result) or (DestSize=CAN_CREATE) then begin
    ICOfh:=Dest;
    BMPih:=ShiftPointer(Dest,sizeof(TICOFileHeader));
    Move(Source^,BMPih^,SourceSize);
    with ICOfh^ do begin
      idReserved:=0;
      idType:=1;
      idCount:=1;
      bWidth:=BMPih^.biWidth;
      bHeight:=BMPih^.biHeight div 2;
      if BMPih^.biBitCount>=8 then bColorCount:=0 else bColorCount:=1 shl BMPih^.biBitCount;
      bReserved:=0;
      wPlanes:=BMPih^.biPlanes;
      wBitCount:=BMPih^.biBitCount;
      dwBytesInRes:=SourceSize;
      dwImageOffset:=$16;
    end;
  end;
end;

function Res2IconGroup(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;ResName:PChar;ResLang:word):integer;
var Text:TStringList;
    Icons:PIcons;
    Header:PIconGroup;
    i:integer;
begin
  if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
  else Text:=nil;
  result:=0;
  try
    Header:=Source;
    Icons:=ShiftPointer(Source,sizeof(TIconGroup));
    result:=AddString(Text,result,'<html><head>');
    result:=AddString(Text,result,'<title>RT_GROUP_ICON :: '+NameToStr(ResName)+' :: '+IntToStr(ResLang)+'</title>');
    result:=AddString(Text,result,'<style type=text/css>');
    result:=AddString(Text,result,'body {margin: 0; font-family: helvetica, sans-serif; font-size: 80%}');
    result:=AddString(Text,result,'td {padding-left: 1 em; padding-right: 1 em; font-size: 80%}');
    result:=AddString(Text,result,'tr {background-color: #EEEEEE; padding-left: 1 em}');
    result:=AddString(Text,result,'h1 {margin: 1 em 2 em 0 em 2 em; padding: 0; font-size: 125%; text-transform: uppercase}');
    result:=AddString(Text,result,'p {margin: 1 em; text-align: justify;}');
    result:=AddString(Text,result,'.head {background-color: #666666; color: #FFFFFF; font-weight: bold}');
    result:=AddString(Text,result,'</style></head><body>');
    result:=AddString(Text,result,'<h1>Resource Information</h1>');
    result:=AddString(Text,result,'<p><table><tr><td width=50%>Resource type</td><td>RT_GROUP_ICON</td></tr>');
    result:=AddString(Text,result,'<tr><td width=50%>Resource name</td><td>'+NameToStr(ResName)+'</td></tr>');
    result:=AddString(Text,result,'<tr><td width=50%>Language</td><td>'+IntToStr(ResLang)+'</td></tr></table></p>');
    result:=AddString(Text,result,'<h1>Resource Content</h1>');
    result:=AddString(Text,result,'<p><table><tr class=head><td>#</td><td>Size</td><td>Bits Per Pixel</td><td>IconID</td></tr>');
    for i:=0 to Header^.idCount-1 do result:=AddString(Text,result,'<tr><td>'+IntToStr(i+1)+'</td><td>'+IntToStr(Icons^[i].bWidth)+'x'+IntToStr(Icons^[i].bHeight)+'</td><td>'+IntToStr(Icons^[i].wBitCount)+' ('+IntToStr(1 shl Icons^[i].wBitCount)+' colors)</td><td>'+IntToStr(Icons^[i].nID)+'</td></tr>');
    result:=AddString(Text,result,'</table></p>');
    result:=AddString(Text,result,'<p><font color=red>NOTE: if you want modify this resource you must use raw binary data not this HTML file.</font></p>');
    result:=AddString(Text,result,'</body><html>');
  finally
    if Assigned(Text) then begin
      if ((DestSize=CAN_CREATE) or (DestSize>=result)) and (result<>UNKNOWN_FORMAT) then begin
        if DestSize=CAN_CREATE then GetMem(Dest,result);
        Move(Pointer(AnsiString(Text.Text))^,Dest^,result);
      end;
      Text.Free;
    end;
  end;
end;

{****  Borland's form  ********************************************************}
function Dfm2Res(Source:pointer;SourceSize:integer;var Dest:pointer;ResName:PChar;ResLang:word):integer;
var s,ObjType,ObjClass:string;
    c:^char;
    i,n,EndChar:integer;
    Txt,Res:TMemoryStream;
begin
  result:=UNKNOWN_FORMAT;
  EndChar:= NativeUInt(ShiftPointer(Source,SourceSize));
  c:=Source; s:='';
  while (integer(c)<EndChar) and (c^<>#13) and (c^<>#10) do begin
    s:=s+c^;
    inc(NativeUInt(c));
  end;
  s:=UpperCase(UpdateString(s));
  i:=pos(' ',s); n:=pos(':',s);
  if ((i=7) or (i=10)) and (n>0) and (i<n) then begin
    ObjType:=copy(s,1,i-1);
    ObjClass:=copy(s,n+1,length(s)-n);
    if ((ObjType='INLINE') or (ObjType='OBJECT') or (ObjType='INHERITED')) and (ObjClass=UpperCase(string(ResName))) then begin
      Txt:=TMemoryStream.Create;
      Res:=TMemoryStream.Create;
      try
        Txt.Write(Source^,SourceSize);
        Txt.Seek(0,soFromBeginning);
        ObjectTextToBinary(Txt,Res);
        result:=Res.Size;
        GetMem(Dest,result); 
        Res.Seek(0,soFromBeginning);
        Res.Read(Dest^,result);
      finally
        Txt.Free;
        Res.Free;
      end;
    end;
  end;
end;

function Res2Dfm(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer):integer;
var Res,Txt:TMemoryStream;
    Txt2,Text:TStringList;
    i,n:integer;
    s,s1:string;
begin
  if (integer(Source^)<>$30465054) then result:=UNKNOWN_FORMAT
  else begin
    Res:=TMemoryStream.Create;
    Txt:=TMemoryStream.Create;
    Txt2:=TStringList.Create;
    if (DestSize=CAN_CREATE) or (DestSize>0) then Text:=TStringList.Create
    else Text:=nil;
    try
      try
        Res.Write(Source^,SourceSize);
        Res.Seek(0,soFromBeginning);
        ObjectBinaryToText(Res,Txt);
        i:=0;
        Txt.Write(i,1);
        Txt2.Text:=String(PAnsiChar(Txt.Memory));
        result:=0;
        while i<Txt2.Count do begin
          s:=Txt2[i];
          s1:=RemoveSpaces(s);
          while (length(s1)>0) and (s1[length(s1)]='+') do begin
            inc(i);
            s1:=RemoveSpaces(Txt2[i]);
            n:=length(s);
            while s[n] in [' ','+'] do dec(n);
            if (s1[1]='''') and (s[n]='''') then begin
              dec(n);
              s1:=copy(s1,2,length(s1)-1);
            end;
            s:=copy(s,1,n)+s1;
          end;
          result:=AddString(Text,result,s);
          inc(i);   
        end;
        if (DestSize=CAN_CREATE) or (DestSize>=result) then begin
          if DestSize=CAN_CREATE then GetMem(Dest,result);
          Move(Pointer(AnsiString(Text.Text))^,Dest^,result);
        end;
      except
        result:=UNKNOWN_FORMAT;
      end;
    finally
      Res.Free;
      Txt.Free;
      Txt2.Free;
      if Assigned(Text) then Text.Free;
    end;
  end;
end;

{******************************************************************************}
{******************************************************************************}
{******************************************************************************}

function FromRes(Source:pointer;SourceSize:integer;var Dest:pointer;DestSize:integer;var Ext:string;ResType,ResName:PChar;ResLang:word):integer;
begin
  Ext:='';
  case integer(ResType) of
    integer(RT_BITMAP):begin;
      result:=Res2Bmp(Source,SourceSize,Dest,DestSize);
      if result<>UNKNOWN_FORMAT then Ext:='.bmp';
    end;
    integer(RT_MENU):begin
      result:=Res2Menu(Source,SourceSize,Dest,DestSize,ResName,ResLang);
      if result<>UNKNOWN_FORMAT then Ext:='.rc';
    end;
    integer(RT_DIALOG):begin
      result:=Res2Dlg(Source,SourceSize,Dest,DEstSize,ResName,ResLang);
      if result<>UNKNOWN_FORMAT then Ext:='.rc';
    end;
    integer(RT_ICON):begin
      result:=Res2Ico(Source,SourceSize,Dest,DestSize);
      if result<>UNKNOWN_FORMAT then Ext:='.ico';
    end;
    integer(RT_CURSOR):begin
      result:=Res2Cur(Source,SourceSize,Dest,DestSize);
      if result<>UNKNOWN_FORMAT then Ext:='.cur';
    end;
    integer(RT_GROUP_ICON):begin
      result:=Res2IconGroup(Source,SourceSize,Dest,DestSize,ResName,ResLang);
      if result<>UNKNOWN_FORMAT then Ext:='.htm';
    end;
    integer(RT_GROUP_CURSOR):begin
      result:=Res2CursorGroup(Source,SourceSize,Dest,DestSize,ResName,ResLang);
      if result<>UNKNOWN_FORMAT then Ext:='.htm';
    end;
    integer(RT_STRING):begin
      result:=Res2Str(Source,SourceSize,Dest,DestSize,ResName,ResLang);
      if result<>UNKNOWN_FORMAT then Ext:='.rc';
    end;
    integer(RT_RCDATA):begin
      result:=Res2Dfm(Source,SourceSize,Dest,DestSize);
      if result<>UNKNOWN_FORMAT then Ext:='.dfm';
    end;
    else result:=UNKNOWN_FORMAT;
  end;
end;

function ToRes(Source:pointer;SourceSize:integer;var Dest:pointer;Ext:string;ResType,ResName:PChar;ResLang:word):integer;
begin
  case integer(ResType) of
    integer(RT_BITMAP):result:=Bmp2Res(Source,SourceSize,Dest);
    integer(RT_MENU):result:=Menu2Res(Source,SourceSize,Dest,ResName,ResLang);
    integer(RT_DIALOG):result:=Dlg2Res(Source,SourceSize,Dest,ResName,ResLang);
    integer(RT_STRING):result:=Str2Res(Source,SourceSize,Dest,ResName,ResLang);
    integer(RT_CURSOR):result:=Cur2Res(Source,SourceSize,Dest);
    integer(RT_ICON):result:=Ico2Res(Source,SourceSize,Dest);
    integer(RT_RCDATA):result:=Dfm2Res(Source,SourceSize,Dest,ResName,ResLang);
    else result:=UNKNOWN_FORMAT;
  end;
end;

end.
