unit resource;

interface

uses Windows, Classes, SysUtils{, utils};

type TResourceItem = record
       ResType:PChar;
       ResName:PChar;
       ResLang:word;
       ResData:pointer;
       ResSize:integer;
       Allocated:boolean;
     end;

const MAXLISTSIZE = (MAXINT div sizeof(TResourceItem))-1;

type TResourceList = array[0..MAXLISTSIZE] of TResourceItem;
     PResourceList = ^TResourceList;

type TResourceData = record
       Data:pointer;
       Size:integer;
     end;

type TResources = class(TPersistent)
     private
       FCopy:boolean;
       FCount:integer;
       FList:PResourceList;
       function CompareItems(ResType1,ResName1:PChar;ResLang1:word;ResType2,ResName2:PChar;ResLang2:word):integer;
       function Find(ResType,ResName:PChar;ResLang:integer;var Index:integer):boolean;
       function GetItem(Index:integer):TResourceItem;
       function GetLanguage(Index:integer):word;
       function GetName(Index:integer):PChar;
       function GetType(Index:integer):PChar;
       function InsertResource(ResType,ResName:PChar;ResLang:word):integer;
       procedure SetItem(Index:integer;Data:pointer;Size:integer);
     protected
       property Items[Index:integer]:TResourceItem read  GetItem;
       function GetResource(Index:integer):TResourceData;
       procedure ExtractResourcesFromModule(H:HMODULE;Copy:boolean);
       procedure SetResource(Index:integer;Value:TResourceData);
     public
       constructor Create;
       destructor Destroy; override;
       function Add(ResType,ResName:PChar;ResLang:word;Data:PChar;Size:integer):integer;
       function IndexOf(ResType,ResName:PChar;ResLang:word):integer;
       procedure Clear;
       procedure Delete(Index:integer);
       procedure UpdateResources(FileName:string;DeleteExisting:boolean);
       property Count:integer read FCount;
       property Languages[Index:integer]:word read GetLanguage;
       property Names[Index:integer]:PChar read GetName;
       property Resources[Index:integer]:TResourceData read GetResource write SetResource; default;
       property Types[Index:integer]:PChar read GetType;
     end;

function NameToInt(str:PChar): NativeInt;
function NameToStr(str:PChar):string;

implementation

{****  Additional functions  **************************************************}

function IsDec(s:string):boolean;
var i:integer;
begin
  if (s='') or (s='-') or (s='+') then result:=false
  else begin
    result:=true;
    if (s[1]='-') or (s[1]='+') then i:=2 else i:=1;
    while (i<=length(s)) and result do begin
      result:=(s[i]>='0') and (s[i]<='9');
      inc(i);
    end;
  end;
end;

function CopyPChar(str:PChar):PChar;
begin
  if IS_INTRESOURCE(str) then result:=str
  else begin
    GetMem(result,StrLen(str) * 2 + 2);
    Move(str^,result^,StrLen(str) * 2 + 2);
  end;
end;

procedure FreePChar(str:PChar);
begin
  if not IS_INTRESOURCE(str) then FreeMem(str,StrLen(str) * 2 + 2);
end;

function NameToInt(str:PChar): NativeInt;
var s:string;
begin
  if IS_INTRESOURCE(str) then result := NativeUInt(str)
  else begin
    s:=string(str);
    if (s[1]='#') and IsDec(copy(s,2,length(s)-1)) then result:=StrToInt(copy(s,2,length(s)-1))
    else result:=-1;
  end;
end;

function NameToStr(str:PChar):string;
var i: NativeInt;
begin
  i:=NameToInt(str);
  if i=-1 then result:=string(str) else result:=IntToStr(i);
end;

function ComparePChar(str1,str2:PChar):integer;
var istr1,istr2: NativeInt;
begin
  istr1:=NativeInt(str1); if (istr1<1) or (istr1>65535) then istr1:=0;
  istr2:=NativeInt(str2); if (istr2<1) or (istr2>65535) then istr2:=0;
  if (istr1=0) and (istr2=0) then result:=AnsiStrIComp(str1,str2)
  else if istr1=0 then result:=-1
  else if istr2=0 then result:=1
  else result:=istr1-istr2;
end;

function EnumResLang(H:HINST;lpType,lpName:PChar;wLanguage:WORD;var Res:TResources):boolean; stdcall;
var HR:HRSRC;
    HG:HGLOBAL;
    Data:pointer;
    Size:integer;
begin
  HR:=FindResourceEx(H,lpType,lpName,wLanguage);
  if HR=0 then RaiseLastOSError;
  HG:=LoadResource(H,HR);
  if HG=0 then RaiseLastOSError;
  Data:=LockResource(HG);
  if Data=nil then RaiseLastOSError;
  Size:=SizeofResource(H,HR);
  if Size=0 then RaiseLastOSError;
  Res.Add(lpType,lpName,wLanguage,Data,Size);
  result:=true;
end;

function EnumResName(H:HINST;lpType,lpName:PChar;LONG_PTR:NativeInt):LongBool; stdcall;
begin
  EnumResourceLanguages(H,lpType,lpName,@EnumResLang,LONG_PTR);
  result:=true;
end;

function EnumResType(H:HINST;lpType:PChar;LONG_PTR:NativeInt):LongBool; stdcall;
begin
  EnumResourceNames(H,lpType,@EnumResName,LONG_PTR);
  result:=true;
end;

{**** TResources functions and procedures  ************************************}
constructor TResources.Create;
begin
  inherited Create;
  FCopy:=true;
  FList:=nil;
  FCount:=0;
end;

destructor TResources.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TResources.Add(ResType,ResName:PChar;ResLang:word;Data:PChar;Size:integer):integer;
begin
  result:=InsertResource(ResType,ResName,ResLang);
  SetItem(result,Data,Size)
end;

function TResources.CompareItems(ResType1,ResName1:PChar;ResLang1:word;ResType2,ResName2:PChar;ResLang2:word):integer;
begin
  result:=ComparePChar(ResType1,ResType2);
  if result=0 then begin
    result:=ComparePChar(ResName1,ResName2);
    if result=0 then result:=ResLang1-ResLang2;
  end;
end;

function TResources.GetItem(Index:integer):TResourceItem;
begin
  if (Index<0) or (Index>=FCount) then raise EListError.Create('Resource index ('+IntToStr(Index)+') is out of bounds (0..'+IntToStr(FCount-1)+').');
  result:=FList^[Index];
end;

function TResources.GetLanguage(Index:integer):word;
begin
  result:=Items[Index].ResLang;
end;

function TResources.GetName(Index:integer):PChar;
begin
  result:=Items[Index].ResName;
end;

function TResources.GetType(Index:integer):PChar;
begin
  result:=Items[Index].ResType;
end;

function TResources.GetResource(Index:integer):TResourceData;
var Item:TResourceItem;
begin
  Item:=Items[Index];
  result.Data:=Item.ResData;
  result.Size:=Item.ResSize;
end;

function TResources.Find(ResType,ResName:PChar;ResLang:integer;var Index:integer):boolean;
var LIndex,RIndex,i:integer;
    Item:TResourceItem;
begin
  result:=false;
  LIndex:=0;
  RIndex:=Count-1;
  if (ResLang<0) or (ResLang>65535) then ResLang:=-1;
  while (LIndex<=RIndex) and (not result) do begin
    Index:=(LIndex+RIndex) div 2;
    Item:=Items[Index];
    if ResLang=-1 then i:=CompareItems(Item.ResType,Item.ResName,0,ResType,ResName,0)
    else i:=CompareItems(Item.ResType,Item.ResName,Item.ResLang,ResType,ResName,ResLang);
    if i<0 then LIndex:=Index+1
    else if i>0 then RIndex:=Index-1
    else result:=true;
  end;
  if not result then Index:=LIndex;
end;

function TResources.IndexOf(ResType,ResName:PChar;ResLang:word):integer;
begin
  if not Find(ResType,ResName,ResLang,result) then result:=-1;
end;

function TResources.InsertResource(ResType,ResName:PChar;ResLang:word):integer;
var Index:integer;
    List:PResourceList;
begin
  if not Find(ResType,ResName,ResLang,Index) then begin
    GetMem(List,sizeof(TResourceItem)*(FCount+1));
    Move(FList^[0],List^[0],sizeof(TResourceItem)*Index);
    Move(FList^[Index],List^[Index+1],sizeof(TResourceItem)*(FCount-Index));
    FreeMem(FList,sizeof(TResourceItem)*FCount);
    FList:=List;
    inc(FCount);
    FList^[Index].ResType:=CopyPChar(ResType);
    FList^[Index].ResName:=CopyPChar(ResName);
    FList^[Index].ResLang:=ResLang;
    FList^[Index].ResData:=nil;
    FList^[Index].ResSize:=0;
    FList^[Index].Allocated:=false;
  end;
  result:=Index;
end;

procedure TResources.Clear;
var i:integer;
begin
  if FList<>nil then begin
    for i:=0 to Count-1 do begin
      if (FList^[i].ResData<>nil) and (FList^[i].Allocated) then FreeMem(FList^[i].ResData,FList^[i].ResSize);
      FreePChar(FList^[i].ResType);
      FreePChar(FList^[i].ResName);
    end;
    FreeMem(FList,sizeof(TResourceItem)*FCount);
    FList:=nil;
    FCount:=0;
  end;
end;

procedure TResources.Delete(Index:integer);
begin
  SetItem(Index,nil,0);
end;

procedure TResources.ExtractResourcesFromModule(H:HMODULE;Copy:boolean);
begin
  FCopy:=Copy;
  try
    EnumResourceTypes(H,@EnumResType, NativeInt(@Self));
  finally
    FCopy:=true;
  end;
end;

procedure TResources.SetItem(Index:integer;Data:pointer;Size:integer);
var p:pointer;
begin
  if (Index<0) or (Index>=FCount) then raise EListError.Create('Resource index ('+IntToStr(Index)+') is out of bounds (0..'+IntToStr(FCount-1)+').');
  if (FList^[Index].ResData<>nil) and (FList^[Index].Allocated) then FreeMem(FList^[Index].ResData,FList^[Index].ResSize);
  if (Data=nil) or (Size=0) then begin
    FList^[Index].ResData:=nil;
    FList^[Index].ResSize:=0;
  end
  else begin
    if FCopy then begin
      GetMem(p,Size);
      Move(Data^,p^,Size);
      FList^[Index].ResData:=p;
      FList^[Index].Allocated:=true;
    end
    else FList^[Index].ResData:=Data;
    FList^[Index].ResSize:=Size;
  end;
end;

procedure TResources.SetResource(Index:integer;Value:TResourceData);
begin
  SetItem(Index,Value.Data,Value.Size);
end;

procedure TResources.UpdateResources(FileName:string;DeleteExisting:boolean);
var H: NativeUInt;
    Item:TResourceItem;
    i:integer;
begin
  H:=BeginUpdateResource(PChar(FileName),DeleteExisting);
  if H=0 then RaiseLastOSError;
  for i:=0 to Count-1 do begin
    Item:=Items[i];
    if Item.ResData=nil then begin
      if not DeleteExisting then UpdateResource(H,Item.ResType,Item.ResName,Item.ResLang,nil,0);
    end
    else UpdateResource(H,Item.ResType,Item.ResName,Item.ResLang,Item.ResData,Item.ResSize);
  end;
  if not EndUpdateResource(H,false) then begin
    RaiseLastOSError;
  end;
end;

end.
