unit resunit;

interface

uses Windows, Classes, SysUtils, wcxhead, resource, convert;

const
  TypesCount = 23;
  Types: array [1 .. TypesCount] of string = ('.RT_CURSOR', '.RT_BITMAP',
    '.RT_ICON', '.RT_MENU', '.RT_DIALOG', '.RT_STRING', '.RT_FONTDIR',
    '.RT_FONT', '.RT_ACCELERATOR', '.RT_RCDATA', '.RT_MESSAGETABLE',
    '.RT_GROUP_CURSOR', '13', '.RT_GROUP_ICON', '15', '.RT_VERSION', '17', '18',
    '.RT_PLUGPLAY', '.RT_VXD', '.RT_ANICURSOR', '.RT_ANIICON', '.RT_HTML');

type
  TArchive = class(TResources)
  private
    FModule: HMODULE;
    FCurrent, FTime: integer;
    FFileName: string;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function Next: boolean;
    procedure Close;
    procedure ExtractResources(FileName: string; Copy: boolean);
    property FileName: string read FFileName;
    property Current: integer read FCurrent;
    property Time: integer read FTime;
  end;

function DeleteFiles(PackedFile:PAnsiChar;DeleteList:PansiChar):integer; stdcall;
function DeleteFilesW(PackedFile:PChar;DeleteList:pointer):integer; stdcall;
function OpenArchive(var ArchiveData:TOpenArchiveData):TArchive; stdcall;
function OpenArchiveW(var ArchiveData:tOpenArchiveDataW):TArchive; stdcall;
function ReadHeader(Arc:TArchive;var HeaderData:THeaderData):integer; stdcall;
function ReadHeaderExW(Arc:TArchive;var HeaderData:THeaderDataExW):integer; stdcall;
function PackFiles(PackedFile,SubPath,SrcPath:PAnsiChar;AddList:pointer;Flags:integer):integer; stdcall;
function PackFilesW(PackedFile,SubPath,SrcPath:PChar;AddList:pointer;Flags:integer):integer; stdcall;

function CanYouHandleThisFile(FileName:PChar):BOOL; stdcall;
function CloseArchive(Arc:TArchive):integer; stdcall;
function GetPackerCaps:integer; stdcall;
function ProcessFile(Arc:TArchive;Operation:integer;DestPath:PAnsiChar;DestName:PAnsiChar):integer; stdcall;
function ProcessFileW(Arc:TArchive;Operation:integer;DestPath:PChar;DestName:PChar):integer; stdcall;

procedure SetProcessDataProc(hArcData:integer;ProcessDataProc:TProcessDataProc); stdcall;
procedure SetProcessDataProcW(hArcData:integer;ProcessDataProcW:TProcessDataProcW); stdcall;
procedure SetChangeVolProc(hArcData:integer;ChangeVolProc:TChangeVolProc); stdcall;
procedure SetChangeVolProcW(hArcData:integer;ChangeVolProcW:TChangeVolProcW); stdcall;

var
  ProcessData:TProcessDataProc;
  ProcessDataW:TProcessDataProcW;

implementation

const E_SUCCESS = 0;
  LOAD_LIBRARY_AS_DATAFILE_EXCLUSIVE = $00000040;
  LOAD_LIBRARY_AS_IMAGE_RESOURCE = $00000020;

function IsDec(s:string):boolean;
var i:integer;
begin
  if (s='') or (s='-') or (s='+') then result:=false
  else begin
    result:=true;
    if (s[1]='-') or (s[1]='+') then i:=2 else i:=1;
    while (i<=length(s)) and result do begin
      result:=(s[i]>='0') and (s[i]<='9');
      inc(i);
    end;
  end;
end;

function CopyPChar(str:PChar):PChar;
begin
  if IS_INTRESOURCE(str) then result:=str
  else begin
    GetMem(result, StrLen(str) * 2 + 2);
    Move(str^,result^,StrLen(str) * 2 + 2);
  end;
end;

function DirectoryExists(const Name:string):boolean;
var Code:Integer;
begin
  Code:=GetFileAttributes(PChar(Name));
  result:=(Code<>-1) and (FILE_ATTRIBUTE_DIRECTORY and Code<>0);
end;

function ForceDirectories(Dir:string):Boolean;
begin
  result:=True;
  if Length(Dir)=0 then raise Exception.Create('Cannot create empty directory.');
  Dir:=ExcludeTrailingPathDelimiter(Dir);
  if (Length(Dir)<3) or DirectoryExists(Dir) or (ExtractFilePath(Dir)=Dir) then Exit;
  result:=ForceDirectories(ExtractFilePath(Dir)) and CreateDir(Dir);
end;

{****** TArchive  *************************************************************}

constructor TArchive.Create;
begin
  inherited;
  FFileName:='';
  FModule:=0;
  FCurrent:=0;
  FTime:=0;
end;

destructor TArchive.Destroy;
begin
  Close;
  inherited;
end;

function TArchive.Next:boolean;
begin
  inc(FCurrent);
  result:=FCurrent<Count;
end;

procedure TArchive.Close;
begin
  if (FModule>0) and (not FreeLibrary(FModule)) then RaiseLastOSError;
  FModule:=0;
end;

procedure TArchive.ExtractResources(FileName:string;Copy:boolean);
begin
  FTime:=FileAge(FileName);
  if CheckWin32Version(6, 0) then
    FModule := Windows.LoadLibraryEx(PWideChar(FileName), 0,
      LOAD_LIBRARY_AS_DATAFILE or LOAD_LIBRARY_AS_IMAGE_RESOURCE)
  else
    FModule := Windows.LoadLibraryEx(PWideChar(FileName), 0,
      LOAD_LIBRARY_AS_DATAFILE or DONT_RESOLVE_DLL_REFERENCES);
  if FModule=0 then RaiseLastOSError;
  ExtractResourcesFromModule(FModule,Copy);
  FCurrent:=-1;
  FFileName:=FileName;
end;

{****** Internal functions  ***************************************************}
function TypeToStr(str:PChar):string;
var i:integer;
begin
  i:=NameToInt(str);
  if i=-1 then result:=string(str)
  else if (i<1) or (i>TypesCount) then result:=IntToStr(i)
  else result:=Types[i];
end;

function StrToName(str:string):PChar;
begin
  if IsDec(str) then result:=PChar(StrToInt(str))
  else result:=CopyPChar(PChar(str));
end;

function StrToType(str:string):PChar;
var s:string;
    i:integer;
begin
  if IsDec(str) then result:=PChar(StrToInt(str))
  else begin
    s:=UpperCase(str);
    i:=1;
    while (i<=TypesCount) and (s<>Types[i]) do inc(i);
    if i>TypesCount then result:=CopyPChar(PChar(str)) else result:=PChar(i);
  end;
end;

function NextPos(Substr,S:string;StartPos:integer):integer;
begin
  result:=StartPos;
  while (result<length(S)-length(Substr)+1) and (Substr<>copy(S,result,length(Substr))) do inc(result);
  if result>=length(S)-length(Substr)+1 then result:=0;
end;

function SplitResourceName(Name:string;var ResType,ResName:PChar;var ResLang:word):boolean;
var i,n:NativeInt;
    p:pointer;
begin
  ResType:=nil;
  ResName:=nil;
  ResLang:=0;
  i:=pos('\',Name); if i=0 then i:=pos('/',Name);
  n:=NextPos('\',Name,i+1); if n=0 then n:=NextPos('/',Name,i+1);
  if (i=0) or (n=0) then result:=false
  else begin
    ResType:=StrToType(copy(Name,1,i-1));
    ResName:=StrToName(copy(Name,n+1,length(Name)-n));
    try
      ResLang:=StrToInt(copy(Name,i+1,n-i-1));
      if not IS_INTRESOURCE(ResName) then begin
        p:=ResName;
        while (byte(p^)<>0) and (byte(p^)<>ord('\')) and (byte(p^)<>ord('/')) do inc(NativeInt(p));
        result:=byte(p^)=0;
      end
      else result:=true;
    except
      result:=false;
    end;
  end;
end;

function ReadOnly:boolean;
var vi:TOSVersionInfo;
begin
  vi.dwOSVersionInfoSize:=sizeof(vi);
  GetVersionEx(vi);
  result:=vi.dwPlatformId<>VER_PLATFORM_WIN32_NT;
end;

function CanConvert:boolean;
begin
  result:=(GetKeyState(VK_SCROLL) and 1)=0;
end;


function GenerateList(ParseChar: Pointer; const Parsed: TStrings): Boolean;
var
  A: PWideChar;
begin
  Result := False;
  if not Assigned(Parsed) then Exit;
  try
    Parsed.Clear;
    if ParseChar = nil then Exit;
    A := ParseChar;
    while (A^ <> #0) do
    begin
      Parsed.Add(string(A));
      Inc(A, StrLen(A)+1);
    end;
    Result := True;
  except
    Result := False;
  end;
end;

function ShowMessage(Text:string):integer;
begin
  if MessageBox(0,PChar(Text+' Do you want to continue ?'),'Error',MB_YESNO or MB_ICONHAND or MB_SETFOREGROUND)=IDNO then result:=E_EABORTED
  else result:=E_SUCCESS;
end;

function MakeDir(str:PChar):string;
begin
  if str=nil then result:=''
  else begin
    result:=string(str);
    if (result<>'') and (result[length(result)]<>'\') and (result[length(result)]<>'/') then result:=result+'\';
  end;
end;

{****** External functions  ***************************************************}

function CanYouHandleThisFile(FileName:PChar):BOOL; stdcall;
var H:HMODULE;
begin
  if CheckWin32Version(6, 0) then
    H := Windows.LoadLibraryEx(FileName, 0,
      LOAD_LIBRARY_AS_DATAFILE or LOAD_LIBRARY_AS_IMAGE_RESOURCE)
  else
    H := Windows.LoadLibraryEx(FileName, 0,
      LOAD_LIBRARY_AS_DATAFILE or DONT_RESOLVE_DLL_REFERENCES);
  if H=0 then result:=false
  else begin
    result:=true;
    FreeLibrary(H);
  end;
end;

function CloseArchive(Arc:TArchive):integer;
begin
  if assigned(Arc) then Arc.Free;
  result:=0;
end;

procedure FreePChar(str:PChar);
begin
  if not IS_INTRESOURCE(str) then FreeMem(str,StrLen(str) * 2 + 2);
end;

function DeleteFiles(PackedFile:PAnsiChar;DeleteList:PAnsiChar):integer;
begin
  result:=E_NOT_SUPPORTED;
end;

function DeleteFilesW(PackedFile:PChar;DeleteList:pointer):integer;
var Files:TStringList;
    Arc:TArchive;
    ResType,ResName:PChar;
    ResLang:word;
    i,n:integer;
begin
  result:=E_SUCCESS;
  if ReadOnly then MessageBox(0,'This function works only under Windows NT/2000/XP.','Error',MB_ICONHAND)
  else begin
    Files:=TStringList.Create;
    Arc:=TArchive.Create;
    ResType:=nil;
    ResName:=nil;
    ResLang:=0;
    try
      try
        Arc.ExtractResources(PackedFile,true);
        Arc.Close;
      except
        result:=E_BAD_ARCHIVE;
      end;
      if result=E_SUCCESS then begin
        GenerateList(DeleteList,Files);
        i:=0;
        while (i<Files.Count) and (result=E_SUCCESS) do begin
          Files[i]:=ChangeFileExt(Files[i],'');
          if Assigned(ProcessDataW) and (ProcessDataW(PChar('Deleting file: '+Files[i]),0)=0) then result:=E_EABORTED;
          if not SplitResourceName(Files[i],ResType,ResName,ResLang) then result:=ShowMessage('"'+Files[i]+'" is not valid resource name.')
          else begin
            n:=Arc.IndexOf(ResType,ResName,ResLang);
            if n=-1 then result:=ShowMessage('Resource "'+Files[i]+'" not found.')
            else Arc.Delete(n);
          end;
          FreePChar(ResType);
          FreePChar(ResName);
          inc(i);
        end;
      end;
      if result=E_SUCCESS then Arc.UpdateResources(PackedFile,true);
    finally
      Files.Free;
      Arc.Free;
    end;
  end;
end;

function GetPackerCaps:integer;
begin
  if ReadOnly then result:=0
  else result:=PK_CAPS_NEW or PK_CAPS_MODIFY or PK_CAPS_MULTIPLE or
    PK_CAPS_DELETE {or PK_CAPS_BY_CONTENT} or PK_CAPS_HIDE;
end;

function OpenArchive(var ArchiveData:TOpenArchiveData):TArchive;
begin
  Result := nil;
end;

function OpenArchiveW(var ArchiveData:tOpenArchiveDataW):TArchive;
begin
  result:=TArchive.Create;
  try
    result.ExtractResources(ArchiveData.ArcName,false);
    ArchiveData.OpenResult:=E_SUCCESS;
  except
    ArchiveData.OpenResult:=E_BAD_DATA;
    result.Free;
    result:=nil;
  end;
end;

function WriteFile2(FileName:string;Data:pointer;Size:cardinal): NativeUInt;
var F:THandle;
    s:string;
    LenToWrite: Cardinal;
begin
  s:=ExtractFileDir(FileName);
  if s<>'' then ForceDirectories(s);
  F:=CreateFile(PChar(FileName),GENERIC_WRITE,0,nil,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
  if F=INVALID_HANDLE_VALUE then result:=INVALID_HANDLE_VALUE
  else begin
    Result := Size;
    LenToWrite := Size;
    if not WriteFile(F,Data^,Size,LenToWrite,nil) then result:=INVALID_HANDLE_VALUE;
    CloseHandle(F);
  end;
end;

function ReadFile2(FileName:string;var Data:pointer): NativeUInt;
var F:THandle;
    Size:cardinal;
    LenToWrite: Cardinal;
begin
  F:=CreateFile(PChar(FileName),GENERIC_READ,FILE_SHARE_READ,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
  if F=INVALID_HANDLE_VALUE then result:=INVALID_HANDLE_VALUE
  else begin
    Size:=GetFileSize(F,nil);
    if Size<>INVALID_FILE_SIZE then begin
      GetMem(Data,Size);
      Result := Size;
      LenToWrite := Size;
      if (not ReadFile(F,Data^,Size,LenToWrite,nil)) or (result<>Size) then begin
        FreeMem
        (Data,Size);
        Data:=nil;
        result:=INVALID_HANDLE_VALUE;
      end;
    end
    else result:=INVALID_HANDLE_VALUE;
    CloseHandle(F);
  end;
end;

function TestFlag(Value,Flag:integer):boolean;
begin
  result:=(Value and Flag)=Flag;
end;

function PackFiles(PackedFile,SubPath,SrcPath:PAnsiChar;AddList:pointer;Flags:integer):integer;
begin
  Result := E_NOT_SUPPORTED;
end;

function PackFilesW(PackedFile,SubPath,SrcPath:PChar;AddList:pointer;Flags:integer):integer;
var Files:TStringList;
    Arc:TArchive;
    FileName,Ext,s:string;
    ResType,ResName:PChar;
    ResLang:word;
    Data,Data2:pointer;
    Size:cardinal;
    i,Size2:integer;
begin
  result:=E_SUCCESS;
  if ReadOnly then MessageBox(0,'This function works only under Windows NT/2000/XP.','Error',MB_ICONHAND)
  else begin
    Files:=TStringList.Create;
    Arc:=TArchive.Create;
    try
      i:=0; GenerateList(AddList,Files);
      while (i<Files.Count) and (result=E_SUCCESS) do begin
        FileName:=MakeDir(SrcPath)+Files[i];
        if Assigned(ProcessDataW) and (ProcessDataW(PChar('Packing file: '+FileName),0)=0) then result:=E_EABORTED;
        if FileExists(FileName) and (result=E_SUCCESS) then begin
          Size:=ReadFile2(FileName,Data);
          if Size=INVALID_HANDLE_VALUE then result:=ShowMessage('Cannot read file "'+FileName+'".')
          else begin
            if TestFlag(Flags,PK_PACK_MOVE_FILES) then begin
              if not DeleteFile(FileName) then result:=ShowMessage('Cannot delete file "'+FileName+'".');
            end;
            Ext:=UpperCase(ExtractFileExt(FileName));
            Files[i]:=MakeDir(SubPath)+ChangeFileExt(Files[i],'');
            if (Ext<>'') and (Ext[1]='.') then Ext:=copy(Ext,2,length(Ext)-1);
            if not SplitResourceName(Files[i],ResType,ResName,ResLang) then result:=ShowMessage('"'+Files[i]+'" is not valid resource name.')
            else begin
              if CanConvert and (Ext<>'') then begin
                Data2:=nil;
                try
                  Size2:=ToRes(Data,Size,Data2,Ext,ResType,ResName,ResLang);
                except
                  on E:Exception do begin
                    if MessageBox(0,PChar('Exception '+E.ClassName+' occur while processing file '+FileName+'.'+#13+E.Message+#13+#13+'Do you want to pack this file as is ?'),'Error',MB_ICONHAND or MB_YESNO)=ID_YES then Size2:=UNKNOWN_FORMAT
                    else Size2:=0;
                  end
                  else Size2:=UNKNOWN_FORMAT;
                end;
                if Size2=UNKNOWN_FORMAT then begin
                  s:=NameToStr(ResName)+'.'+Ext;
                  FreePChar(ResName);
                  ResName:=CopyPChar(PChar(s));
                  Arc.Add(ResType,ResName,ResLang,Data,Size)
                end
                else if Size2>0 then begin
                  Arc.Add(ResType,ResName,ResLang,Data2,Size2);
                  FreeMem(Data2,Size2);
                end;
              end
              else Arc.Add(ResType,ResName,ResLang,Data,Size);
            end;
            FreePChar(ResType);
            FreePChar(ResName);
            if Data<>nil then begin
              FreeMem(Data,Size);
              Data:=nil;
            end;
          end;
        end;
        inc(i);
      end;
      if result=E_SUCCESS then Arc.UpdateResources(PackedFile,false);
    finally
      Files.Free;
      Arc.Free;
    end;
  end;
end;

function ProcessFile(Arc:TArchive;Operation:integer;DestPath:PAnsiChar;DestName:PAnsiChar):integer;
begin
  Result := E_NOT_SUPPORTED;
end;

function ProcessFileW(Arc:TArchive;Operation:integer;DestPath:PChar;DestName:PChar):integer;
var FileName,Ext:string;
    Data:pointer;
    Size:integer;
begin
  result:=E_SUCCESS;
  if Operation=PK_EXTRACT then begin
    if DestName=nil then FileName:=string(DestPath)
    else if DestPath=nil then FileName:=string(DestName)
    else FileName:=string(DestPath)+'\'+string(DestName);
    Data:=nil; Ext:=''; Size:=UNKNOWN_FORMAT;
    if CanConvert then Size:=FromRes(Arc[Arc.Current].Data,Arc[Arc.Current].Size,Data,CAN_CREATE,Ext,Arc.Types[Arc.Current],Arc.Names[Arc.Current],Arc.Languages[Arc.Current]);
    if Size=UNKNOWN_FORMAT then begin
      if WriteFile2(FileName,Arc[Arc.Current].Data,Arc[Arc.Current].Size)=INVALID_HANDLE_VALUE then result:=E_EWRITE;
    end
    else begin
      if WriteFile2(FileName,Data,Size)=INVALID_HANDLE_VALUE then result:=E_EWRITE;
    end;
    FreeMem(Data);
  end;
end;
{
function StrToBuffer(const s:string;var Buffer:array of char;SizeOfBuffer:integer):PChar;
var i:integer;
begin
  i:=1;
  if SizeOfBuffer=-1 then SizeOfBuffer:=length(s)
  else ZeroMemory(@Buffer,SizeOfBuffer);
  while (i<=length(s)) and (i<=SizeOfBuffer) do begin
    Buffer[i-1]:=s[i];
    inc(i);
  end;
  result:=@Buffer;
end;
}
function StrToBuffer(const s:string;var Buffer:array of char;SizeOfBuffer:integer):PChar;
begin
  StrLCopy(Buffer, PChar(s), SizeOfBuffer);
  result:=@Buffer;
end;

function ReadHeader(Arc:TArchive;var HeaderData:THeaderData):integer;
begin
  Result := E_NOT_SUPPORTED;
end;

function ReadHeaderExW(Arc:TArchive;var HeaderData:THeaderDataExW):integer; stdcall;
var Ext:string;
    Data:pointer;
    Size:integer;
begin
  if Arc.Next then begin
    result:=E_SUCCESS;
    with HeaderData do begin
      try
        Data:=nil; Size:=-1;
        if CanConvert then Size:=FromRes(Arc[Arc.Current].Data,Arc[Arc.Current].Size,Data,CAN_CREATE,Ext,Arc.Types[Arc.Current],Arc.Names[Arc.Current],Arc.Languages[Arc.Current]);
        if Size=UNKNOWN_FORMAT then Size:=Arc[Arc.Current].Size;
        StrToBuffer(Arc.FileName,ArcNameW, Length(ArcNameW) - 1);
        StrToBuffer(TypeToStr(Arc.Types[Arc.Current])+'\'+IntToStr(Arc.Languages[Arc.Current])+'\'+NameToStr(Arc.Names[Arc.Current])+Ext, FileNameW , Length(FileNameW) - 1);
        Flags:=0;
        PackSize:=Size;
        UnpSize:=PackSize;
        HostOS:=0;
        FileCRC:=0;
        FileTime:=Arc.Time;
        UnpVer:=0;
        Method:=0;
        FileAttr:=faArchive;
        CmtBuf:=nil;
        CmtSize:=0;
        cmtState:=0;
        FreeMem(Data);
      except
        result:=E_BAD_ARCHIVE;
      end;
    end;
  end
  else result:=E_END_ARCHIVE;
end;

procedure SetProcessDataProc(hArcData:integer;ProcessDataProc:TProcessDataProc);
begin
  ProcessData:=ProcessDataProc;
end;

procedure SetProcessDataProcW(hArcData:integer; ProcessDataProcW:TProcessDataProcW);
begin
  ProcessDataW:=ProcessDataProcW;
end;

procedure SetChangeVolProc(hArcData:integer;ChangeVolProc:TChangeVolProc);
begin

end;

procedure SetChangeVolProcW(hArcData:integer;ChangeVolProcW:TChangeVolProcW);
begin

end;

end.
